@@ -25,16 +25,13 @@
 #include <stdlib.h>
 #include <windows.h>
 #include <io.h>
-#include <fcntl.h>
 #include <signal.h>		/* my_system */
 #include <limits.h>		/* INT_MAX */
 #include <errno.h>
-#include <sys/time.h>		/* select: timeout */
-#include <sys/param.h>
 #include <sys/stat.h>
 #include <stdarg.h>
 #include <process.h>
-#include "../src/fs.h"
+#include "../src/global.h"
 #include "../src/util.h"
 #include "util_win32.h"
 
@@ -65,27 +62,30 @@ static int old_error;		/* File descriptor of old standard error */
 void open_error_pipe (void)
 {
     if (pipe (error_pipe) < 0){
-	message (0, " Warning ", " Pipe failed ");
+	message (0, "Warning", " Pipe failed ");
     }
     old_error = dup (2);
     if(old_error < 0 || close(2) || dup (error_pipe[1]) != 2){
-	message (0, " Warning ", " Dup failed ");
+	message (0, "Warning", " Dup failed ");
 	close (error_pipe[0]);
 	close (error_pipe[1]);
     }
     close (error_pipe[1]);
 }
 
-void close_error_pipe (int error, char *text)
+/*
+ * Returns true if an error was displayed
+ */
+int close_error_pipe (int error, char *text)
 {
     char *title;
     char msg[MAX_PIPE_SIZE];
     int len = 0;
 
     if (error)
-	title = " Error ";
+	title = "Error";
     else
-	title = " Warning ";
+	title = "Warning";
     if (old_error >= 0){
 	close (2);
 	dup (old_error);
@@ -97,16 +97,17 @@ void close_error_pipe (int error, char *text)
 	close (error_pipe[0]);
     }
     if (error < 0)
-	return;		/* Just ignore error message */
+	return 0;		/* Just ignore error message */
     if (text == NULL){
-	if (len == 0) return;	/* Nothing to show */
+	if (len == 0) return 0;	/* Nothing to show */
 
 	/* Show message from pipe */
-	message (error, title, msg);
+	message (error, title, "%s", msg);
     } else {
 	/* Show given text and possible message from pipe */
 	message (error, title, " %s \n %s ", text, msg);
     }
+    return 1;
 }
 
 void check_error_pipe (void)
@@ -118,7 +119,7 @@ void check_error_pipe (void)
 	{
 	    int rvalue;
 
-	    rvalue = -1; // read (error_pipe[0], error + len, 1);
+	    rvalue = -1; /* read (error_pipe[0], error + len, 1); */
 	    if (rvalue <= 0)
 		break;
 	    len ++;
@@ -127,7 +128,7 @@ void check_error_pipe (void)
 	close (error_pipe[0]);
     }
     if (len > 0)
-        message (0, " Warning ", error);
+        message (0, "Warning", "%s", error);
 }
 
 int my_system (int as_shell_command, const char *shell, const char *command)
@@ -188,13 +189,13 @@ char* get_default_shell()
 	return "command.com";
 }
 
-char *tilde_expand (char *directory)
+char *tilde_expand (const char *directory)
 {
     return strdup (directory);
 }
 
 /* sleep: Call Windows API.
-	  Can't do simple define. That would need <windows.h> in every source
+	  Cannot do simple define. That would need <windows.h> in every source
 */
 #ifndef __EMX__
 void sleep(unsigned long dwMiliSecs)
@@ -241,10 +242,6 @@ char *canonicalize_pathname (char *path)
 	    i = start + 1;
 	}
 
-        /* Handle backquoted `/'. */
-        if (start > 0 && path[start - 1] == '\\')
-	    continue;
-
         /* Check for trailing `/'. */
         if (start && !path[i]) {
 	zero_last:
@@ -597,23 +594,23 @@ int mc_doublepopen (int inhandle, int inlen, pid_t *the_pid, char *command, ...)
 
     pid_t pid;
 
-	// Create the pipes
+	/* Create the pipes */
 	if(_pipe(pipe0, 8192, O_BINARY | O_NOINHERIT) == -1)
    	    exit (1);
 	if(_pipe(pipe1, 8192, O_BINARY | O_NOINHERIT) == -1)
    	    exit (1);
-	// Duplicate stdin/stdout handles (next line will close original)
+	/* Duplicate stdin/stdout handles (next line will close original) */
 	std_sav[0] = _dup(_fileno(stdin));
 	std_sav[1] = _dup(_fileno(stdout));
-	// Duplicate read end of pipe0 to stdin handle
+	/* Duplicate read end of pipe0 to stdin handle */
 	if(_dup2(pipe0[0], _fileno(stdin)) != 0)
    	    exit (1);
-	// Duplicate write end of pipe1 to stdout handle
+	/* Duplicate write end of pipe1 to stdout handle */
 	if(_dup2(pipe1[1], _fileno(stdout)) != 0)
    	    exit (1);
-	// Close original read end of pipe0
+	/* Close original read end of pipe0 */
 	close(pipe0[0]);
-	// Close original write end of pipe1
+	/* Close original write end of pipe1 */
 	close(pipe1[1]);
 
 	va_start (ap, command);
@@ -624,17 +621,17 @@ int mc_doublepopen (int inhandle, int inlen, pid_t *the_pid, char *command, ...)
 		break;
 	}
 	va_end (ap);
-	// Spawn process
-	pid = spawnvp(P_NOWAIT,command, args);// argv[1], (const char* const*)&argv[1]);
+	/* Spawn process */
+	pid = spawnvp(P_NOWAIT,command, args);/* argv[1], (const char* const*)&argv[1]); */
 	if(!pid)
    	    exit (1);
-	// Duplicate copy of original stdin back into stdin
+	/* Duplicate copy of original stdin back into stdin */
 	if(_dup2(std_sav[0], _fileno(stdin)) != 0)
    	    exit (1);
-	// Duplicate copy of original stdout back into stdout
+	/* Duplicate copy of original stdout back into stdout */
 	if(_dup2(std_sav[1], _fileno(stdout)) != 0)
    	    exit (1);
-	// Close duplicate copy of original stdout  and stdin    
+	/* Close duplicate copy of original stdout  and stdin */ 
 	close(std_sav[0]);
 	close(std_sav[1]);
 
