@@ -1,8 +1,6 @@
 #ifndef __BOXES_H
 #define __BOXES_H
 
-extern char *user_format [];
-
 int     display_box      (WPanel *p, char **user, char **mini,
 			  int *use_msformat, int num);
 sortfn *sort_box         (sortfn *sort_fn, int *reverse,
@@ -14,5 +12,6 @@ void    jobs_cmd         (void);
 char   *cd_dialog        (void);
 void    symlink_dialog   (char *existing, char *new,
 			  char **ret_existing, char **ret_new);
+char   *tree_box         (char *current_dir);
 
 #endif
