@@ -17,34 +17,23 @@
 
 #include <config.h>
 #define DIR_H_INCLUDE_HANDLE_DIRENT
-#include "tty.h"
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <sys/stat.h>
-#include "x.h"
+
 #include "global.h"
+#include "tty.h"
 #include "dir.h"
-#include "dialog.h"
+#include "wtools.h"
 #include "tree.h"
-#include "../vfs/vfs.h"
 
-/* "$Id: dir.c,v 1.23 2001/01/27 05:51:27 proskin Exp $" */
-
-/*
- * If true show files starting with a dot.
- *
- * In GNOME we default to not showing them.
- */
-#ifdef HAVE_GNOME
-int show_dot_files = 0;
-#else
+/* If true show files starting with a dot */
 int show_dot_files = 1;
-#endif
 
 /* If true show files ending in ~ */
-int show_backups = 0;
+int show_backups = 1;
 
 /* If false then directories are shown separately from files */
 int mix_all_files = 0;
@@ -76,14 +65,44 @@ sort_orders_t sort_orders [SORT_TYPES_TOTAL] = {
     { N_("&Group"),       sort_group }
 };
 
-/*
- * FIXME: strcoll() is not case sensitive in some locales, including en_US.
- * The user gets a "choice" between two case insensitive sorts.
- * Ideally, the user should be able to select between all three functions -
- * strcoll() (if available), strcmp() and g_strcasecmp().
- */
 #ifdef HAVE_STRCOLL
-#define string_sortcomp(a,b) (case_sensitive ? strcoll (a,b) : g_strcasecmp (a,b))
+/*
+ * g_strcasecmp() doesn't work well in some locales because it relies on
+ * the locale-specific toupper().  On the other hand, strcoll() is case
+ * sensitive in the "C" and "POSIX" locales, unlike other locales.
+ * Solution: always use strcmp() for case sensitive sort.  For case
+ * insensitive sort use strcoll() if it's case insensitive for ASCII and
+ * g_strcasecmp() otherwise.
+ */
+typedef enum {
+    STRCOLL_NO,
+    STRCOLL_YES,
+    STRCOLL_TEST	
+} strcoll_status;
+
+static int string_sortcomp (char *str1, char *str2)
+{
+    static strcoll_status use_strcoll = STRCOLL_TEST;
+
+    if (case_sensitive) {
+	return strcmp (str1, str2);
+    }
+
+    /* Initialize use_strcoll once.  */
+    if (use_strcoll == STRCOLL_TEST) {
+	/* Only use strcoll() if it considers "B" between "a" and "c".  */
+	if (strcoll ("a", "B") * strcoll ("B", "c") > 0) {
+	    use_strcoll = STRCOLL_YES;
+	} else {
+	    use_strcoll = STRCOLL_NO;
+	}
+    }
+
+    if (use_strcoll == STRCOLL_NO)
+	return g_strcasecmp (str1, str2);
+    else
+	return strcoll (str1, str2);
+}
 #else
 #define string_sortcomp(a,b) (case_sensitive ? strcmp (a,b) : g_strcasecmp (a,b))
 #endif
@@ -201,10 +220,10 @@ sort_size (const file_entry *a, const file_entry *b)
     int ad = MY_ISDIR (a);
     int bd = MY_ISDIR (b);
 
-    if (ad == bd || mix_all_files)
-	return (b->buf.st_size - a->buf.st_size) * reverse;
-    else
-	return bd-ad;
+    if (ad != bd && !mix_all_files)
+	return bd - ad;
+
+    return (2 * (b->buf.st_size > a->buf.st_size) - 1) * reverse;
 }
 
 int
@@ -253,7 +272,7 @@ file_type_to_num (const file_entry *fe)
     if (S_ISLNK (s->st_mode)){
 	if (fe->f.link_to_dir)
 	    return 1;
-	if (fe->f.stalled_link)
+	if (fe->f.stale_link)
 	    return 2;
 	else
 	    return 3;
@@ -285,10 +304,12 @@ void
 do_sort (dir_list *list, sortfn *sort, int top, int reverse_f, int case_sensitive_f)
 {
     int i;
+    int dot_dot_found = 0;
     file_entry tmp_fe;
 
     for (i = 0; i < top + 1; i++) {             /* put ".." first in list */
 	if (!strcmp (list->list [i].fname, "..")) {
+	    dot_dot_found = 1;
             if (i > 0) {                        /* swap [i] and [0] */
                 memcpy (&tmp_fe, &(list->list [0]), sizeof (file_entry));
                 memcpy (&(list->list [0]), &(list->list [i]), sizeof (file_entry));
@@ -300,7 +321,8 @@ do_sort (dir_list *list, sortfn *sort, int top, int reverse_f, int case_sensitiv
 
     reverse = reverse_f ? -1 : 1;
     case_sensitive = case_sensitive_f;
-    qsort (&(list->list) [1], top, sizeof (file_entry), sort);
+    qsort (&(list->list) [dot_dot_found],
+	   top + 1 - dot_dot_found, sizeof (file_entry), sort);
 }
 
 void
@@ -317,10 +339,6 @@ clean_dir (dir_list *list, int count)
 static int
 add_dotdot_to_list (dir_list *list, int index)
 {
-    char buffer [MC_MAXPATHLEN + MC_MAXPATHLEN];
-    char *p;
-    int i = 0;
-
     /* Need to grow the *list? */
     if (index == list->size) {
 	list->list = g_realloc (list->list, sizeof (file_entry) *
@@ -330,40 +348,14 @@ add_dotdot_to_list (dir_list *list, int index)
 	list->size += RESIZE_STEPS;
     }
 
+    memset (&(list->list) [index], 0, sizeof(file_entry));
     (list->list) [index].fnamelen = 2;
     (list->list) [index].fname = g_strdup ("..");
     (list->list) [index].f.link_to_dir = 0;
-    (list->list) [index].f.stalled_link = 0;
+    (list->list) [index].f.stale_link = 0;
     (list->list) [index].f.dir_size_computed = 0;
-
-    /* FIXME: We need to get the panel definition! to use file_mark */
     (list->list) [index].f.marked = 0;
-    mc_get_current_wd (buffer, sizeof (buffer) - 1 );
-    if (buffer [strlen (buffer) - 1] == PATH_SEP)
-    	buffer [strlen (buffer) - 1] = 0;
-    for (;;) {
-    	strcat (buffer, PATH_SEP_STR "..");
-        p = vfs_canon (buffer);
-        if (mc_stat (p, &((list->list) [index].buf)) != -1){
-	    g_free (p);
-            break;
-	}
-        i = 1;
-        if (!strcmp (p, PATH_SEP_STR)){
-	    g_free (p);
-            return 1;
-	}
-	strcpy (buffer, p);
-	g_free (p);
-    }
-
-/* Commented out to preserve a usable '..'. What's the purpose of this
- * three lines? (Norbert) */
-#if 0
-    if (i) { /* So there is bogus information on the .. directory's stat */
-        (list->list) [index].buf.st_mode &= ~0444;
-    }
-#endif
+    (list->list) [index].buf.st_mode = 040755;
     return 1;
 }
 
@@ -379,22 +371,23 @@ set_zero_dir (dir_list *list)
 int
 handle_dirent (dir_list *list, char *filter, struct dirent *dp,
 	       struct stat *buf1, int next_free, int *link_to_dir,
-	       int *stalled_link)
+	       int *stale_link)
 {
-    if (dp->d_name [0] == '.' && dp->d_name [1] == 0)
+    if (dp->d_name[0] == '.' && dp->d_name[1] == 0)
 	return 0;
-    if (!show_dot_files){
-	if (dp->d_name [0] == '.'){
-	    if (!(dp->d_name [1] == 0))
-		if (!(dp->d_name [1] == '.' && NLENGTH (dp) == 2))
-		    return 0;
-	}
-    }
-    if (!show_backups && dp->d_name [NLENGTH (dp)-1] == '~')
+    if (dp->d_name[0] == '.' && dp->d_name[1] == '.' && dp->d_name[2] == 0)
+	return 0;
+    if (!show_dot_files && (dp->d_name[0] == '.'))
+	return 0;
+    if (!show_backups && dp->d_name[NLENGTH (dp) - 1] == '~')
 	return 0;
     if (mc_lstat (dp->d_name, buf1) == -1) {
-	message(1, _(" Error "), _("File '%s' exists but can not be stat-ed: %s"), dp->d_name, strerror(errno));
-        return 0;
+	/*
+	 * lstat() fails - such entries should be identified by
+	 * buf1->st_mode being 0.
+	 * It happens on QNX Neutrino for /fs/cd0 if no CD is inserted.
+	 */
+	memset (buf1, 0, sizeof (*buf1));
     }
 
     if (S_ISDIR (buf1->st_mode))
@@ -402,22 +395,23 @@ handle_dirent (dir_list *list, char *filter, struct dirent *dp,
 
     /* A link to a file or a directory? */
     *link_to_dir = 0;
-    *stalled_link = 0;
-    if (S_ISLNK(buf1->st_mode)){
+    *stale_link = 0;
+    if (S_ISLNK (buf1->st_mode)) {
 	struct stat buf2;
 	if (!mc_stat (dp->d_name, &buf2))
-	    *link_to_dir = S_ISDIR(buf2.st_mode) != 0;
+	    *link_to_dir = S_ISDIR (buf2.st_mode) != 0;
 	else
-	    *stalled_link = 1;
+	    *stale_link = 1;
     }
-    if (!(S_ISDIR(buf1->st_mode) || *link_to_dir) && filter &&
-	!regexp_match (filter, dp->d_name, match_file))
+    if (!(S_ISDIR (buf1->st_mode) || *link_to_dir) && filter
+	&& !regexp_match (filter, dp->d_name, match_file))
 	return 0;
 
     /* Need to grow the *list? */
-    if (next_free == list->size){
-	list->list = g_realloc (list->list, sizeof (file_entry) *
-			      (list->size + RESIZE_STEPS));
+    if (next_free == list->size) {
+	list->list =
+	    g_realloc (list->list,
+		       sizeof (file_entry) * (list->size + RESIZE_STEPS));
 	if (!list->list)
 	    return -1;
 	list->size += RESIZE_STEPS;
@@ -433,10 +427,12 @@ handle_dirent (dir_list *list, char *filter, struct dirent *dp,
 int
 handle_path (dir_list *list, char *path,
 	     struct stat *buf1, int next_free, int *link_to_dir,
-	     int *stalled_link)
+	     int *stale_link)
 {
     if (path [0] == '.' && path [1] == 0)
 	return 0;
+    if (path [0] == '.' && path [1] == '.' && path [2] == 0)
+	return 0;
     if (mc_lstat (path, buf1) == -1)
         return 0;
 
@@ -445,13 +441,13 @@ handle_path (dir_list *list, char *path,
 
     /* A link to a file or a directory? */
     *link_to_dir = 0;
-    *stalled_link = 0;
+    *stale_link = 0;
     if (S_ISLNK(buf1->st_mode)){
 	struct stat buf2;
 	if (!mc_stat (path, &buf2))
 	    *link_to_dir = S_ISDIR(buf2.st_mode) != 0;
 	else
-	    *stalled_link = 1;
+	    *stale_link = 1;
     }
 
     /* Need to grow the *list? */
@@ -466,51 +462,54 @@ handle_path (dir_list *list, char *path,
 }
 
 int
-do_load_dir (dir_list *list, sortfn *sort, int reverse, int case_sensitive, char *filter)
+do_load_dir (dir_list *list, sortfn *sort, int reverse, int case_sensitive,
+	     char *filter)
 {
-    DIR           *dirp;
+    DIR *dirp;
     struct dirent *dp;
-    int           status, link_to_dir, stalled_link;
-    int           next_free = 0;
-    struct stat   buf;
-    int dotdot_found = 0;
+    int status, link_to_dir, stale_link;
+    int next_free = 0;
+    struct stat buf;
 
     tree_store_start_check_cwd ();
 
     dirp = mc_opendir (".");
-    if (!dirp){
+    if (!dirp) {
+	message (1, MSG_ERROR, _("Cannot read directory contents"));
 	tree_store_end_check ();
 	return set_zero_dir (list);
     }
-    for (dp = mc_readdir (dirp); dp; dp = mc_readdir (dirp)){
-	status = handle_dirent (list, filter, dp, &buf, next_free, &link_to_dir,
-	    &stalled_link);
+    for (dp = mc_readdir (dirp); dp; dp = mc_readdir (dirp)) {
+	status =
+	    handle_dirent (list, filter, dp, &buf, next_free, &link_to_dir,
+			   &stale_link);
 	if (status == 0)
 	    continue;
-	if (status == -1){
+	if (status == -1) {
 	    tree_store_end_check ();
+	    mc_closedir (dirp);
 	    return next_free;
 	}
-	list->list [next_free].fnamelen = NLENGTH (dp);
-	list->list [next_free].fname = g_strdup (dp->d_name);
-	list->list [next_free].f.marked = 0;
-	list->list [next_free].f.link_to_dir = link_to_dir;
-	list->list [next_free].f.stalled_link = stalled_link;
-        list->list [next_free].f.dir_size_computed = 0;
-	list->list [next_free].buf = buf;
-	if (strcmp (dp->d_name, ".." ) == 0)
-	    dotdot_found = 1;
+	list->list[next_free].fnamelen = NLENGTH (dp);
+	list->list[next_free].fname = g_strdup (dp->d_name);
+	list->list[next_free].f.marked = 0;
+	list->list[next_free].f.link_to_dir = link_to_dir;
+	list->list[next_free].f.stale_link = stale_link;
+	list->list[next_free].f.dir_size_computed = 0;
+	list->list[next_free].buf = buf;
 	next_free++;
 	if (!(next_free % 32))
 	    rotate_dash ();
     }
 
     if (next_free) {
-	if (!dotdot_found)
+	/* Add ".." except the root directory */
+	if (strcmp (vfs_canon ("."), "/") != 0)
 	    add_dotdot_to_list (list, next_free++);
-	do_sort (list, sort, next_free-1, reverse, case_sensitive);
+	do_sort (list, sort, next_free - 1, reverse, case_sensitive);
     } else {
 	tree_store_end_check ();
+	mc_closedir (dirp);
 	return set_zero_dir (list);
     }
 
@@ -568,115 +567,101 @@ alloc_dir_copy (int size)
 
 /* If filter is null, then it is a match */
 int
-do_reload_dir (dir_list *list, sortfn *sort, int count, int rev,
+do_reload_dir (dir_list * list, sortfn * sort, int count, int rev,
 	       int case_sensitive, char *filter)
 {
-    DIR           *dirp;
+    DIR *dirp;
     struct dirent *dp;
-    int           next_free = 0;
-    int           i, found, status, link_to_dir, stalled_link;
-    struct stat   buf;
-    int		  tmp_len;  /* For optimisation */
-    int 	  dotdot_found = 0;
+    int next_free = 0;
+    int i, status, link_to_dir, stale_link;
+    struct stat buf;
+    int marked_cnt;
+    GHashTable *marked_files = g_hash_table_new (g_str_hash, g_str_equal);
 
     tree_store_start_check_cwd ();
     dirp = mc_opendir (".");
     if (!dirp) {
- 	clean_dir (list, count);
+	message (1, MSG_ERROR, _("Cannot read directory contents"));
+	clean_dir (list, count);
 	tree_store_end_check ();
 	return set_zero_dir (list);
     }
 
     alloc_dir_copy (list->size);
-    for (i = 0; i < count; i++){
-	dir_copy.list [i].fnamelen = list->list [i].fnamelen;
-	dir_copy.list [i].fname =    list->list [i].fname;
-	dir_copy.list [i].f.marked = list->list [i].f.marked;
-        dir_copy.list [i].f.dir_size_computed = list->list [i].f.dir_size_computed;
-	dir_copy.list [i].f.link_to_dir = list->list [i].f.link_to_dir;
-	dir_copy.list [i].f.stalled_link = list->list [i].f.stalled_link;
+    for (marked_cnt = i = 0; i < count; i++) {
+	dir_copy.list[i].fnamelen = list->list[i].fnamelen;
+	dir_copy.list[i].fname = list->list[i].fname;
+	dir_copy.list[i].f.marked = list->list[i].f.marked;
+	dir_copy.list[i].f.dir_size_computed =
+	    list->list[i].f.dir_size_computed;
+	dir_copy.list[i].f.link_to_dir = list->list[i].f.link_to_dir;
+	dir_copy.list[i].f.stale_link = list->list[i].f.stale_link;
+	if (list->list[i].f.marked) {
+	    g_hash_table_insert (marked_files, dir_copy.list[i].fname,
+				 &dir_copy.list[i]);
+	    marked_cnt++;
+	}
     }
 
-    for (dp = mc_readdir (dirp); dp; dp = mc_readdir (dirp)){
-	status = handle_dirent (list, filter, dp, &buf, next_free, &link_to_dir,
-	    &stalled_link);
+    for (dp = mc_readdir (dirp); dp; dp = mc_readdir (dirp)) {
+	status =
+	    handle_dirent (list, filter, dp, &buf, next_free, &link_to_dir,
+			   &stale_link);
 	if (status == 0)
 	    continue;
 	if (status == -1) {
 	    mc_closedir (dirp);
 	    /* Norbert (Feb 12, 1997):
-	     Just in case someone finds this memory leak:
-	     -1 means big trouble (at the moment no memory left),
-	     I don't bother with further cleanup because if one gets to
-	     this point he will have more problems than a few memory
-	     leaks and because one 'clean_dir' would not be enough (and
-	     because I don't want to spent the time to make it working,
-             IMHO it's not worthwhile).
-	    clean_dir (&dir_copy, count);
-             */
+	       Just in case someone finds this memory leak:
+	       -1 means big trouble (at the moment no memory left),
+	       I don't bother with further cleanup because if one gets to
+	       this point he will have more problems than a few memory
+	       leaks and because one 'clean_dir' would not be enough (and
+	       because I don't want to spent the time to make it working,
+	       IMHO it's not worthwhile).
+	       clean_dir (&dir_copy, count);
+	     */
 	    tree_store_end_check ();
 	    return next_free;
 	}
 
-	tmp_len = NLENGTH (dp);
-	for (found = i = 0; i < count; i++)
-	    if (tmp_len == dir_copy.list [i].fnamelen
-		&& !strcmp (dp->d_name, dir_copy.list [i].fname)){
-		list->list [next_free].f.marked = dir_copy.list [i].f.marked;
-                found = 1;
-		break;
+	list->list[next_free].f.marked = 0;
+
+	/*
+	 * If we have marked files in the copy, scan through the copy
+	 * to find matching file.  Decrease number of remaining marks if
+	 * we copied one.
+	 */
+	if (marked_cnt > 0) {
+	    file_entry *p;
+	    if (NULL !=
+		(p = g_hash_table_lookup (marked_files, dp->d_name))) {
+		list->list[next_free].f.marked = 1;
+		marked_cnt--;
 	    }
+	}
 
-	if (!found)
-	    list->list [next_free].f.marked = 0;
-
-	list->list [next_free].fnamelen = tmp_len;
-	list->list [next_free].fname = g_strdup (dp->d_name);
-	list->list [next_free].f.link_to_dir = link_to_dir;
-	list->list [next_free].f.stalled_link = stalled_link;
-        list->list [next_free].f.dir_size_computed = 0;
-	list->list [next_free].buf = buf;
-	if (strcmp (dp->d_name, ".." ) == 0)
-	    dotdot_found = 1;
+	list->list[next_free].fnamelen = NLENGTH (dp);
+	list->list[next_free].fname = g_strdup (dp->d_name);
+	list->list[next_free].f.link_to_dir = link_to_dir;
+	list->list[next_free].f.stale_link = stale_link;
+	list->list[next_free].f.dir_size_computed = 0;
+	list->list[next_free].buf = buf;
 	next_free++;
 	if (!(next_free % 16))
 	    rotate_dash ();
     }
     mc_closedir (dirp);
     tree_store_end_check ();
+    g_hash_table_destroy (marked_files);
     if (next_free) {
-	if (!dotdot_found)
+	/* Add ".." except the root directory */
+	if (strcmp (vfs_canon ("."), "/") != 0)
 	    add_dotdot_to_list (list, next_free++);
-	do_sort (list, sort, next_free-1, rev, case_sensitive);
-    }
-    else
+	do_sort (list, sort, next_free - 1, rev, case_sensitive);
+    } else
 	next_free = set_zero_dir (list);
     clean_dir (&dir_copy, count);
     return next_free;
 }
 
-char *
-sort_type_to_name (sortfn *sort_fn)
-{
-    int i;
-
-    for (i = 0; i < SORT_TYPES; i++)
-	if ((sortfn *) (sort_orders [i].sort_fn) == sort_fn)
-	    return _(sort_orders [i].sort_name);
-
-    return _("Unknown");
-}
-
-sortfn *
-sort_name_to_type (char *sname)
-{
-    int i;
-
-    for (i = 0; i < SORT_TYPES; i++)
-	if ( g_strcasecmp (sort_orders [i].sort_name, sname) == 0)
-	    return (sortfn *) sort_orders [i].sort_fn;
-
-    /* default case */
-    return (sortfn *) sort_name;
-}
-
