@@ -37,40 +37,39 @@
 */
 
 #include <config.h>
-#include "tty.h"
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <errno.h>
+
 #include "global.h"
+#include "tty.h"
 #include "color.h"
 #include "dialog.h"
 #include "win.h"
 #include "mouse.h"
 #include "key.h"	/* For mi_getch() */
 #include "help.h"
-#include "layout.h"		/* keybar_visible */
-#include "x.h"
 #include "dlg.h"		/* For Dlg_head */
 #include "widget.h"		/* For Widget */
-
-#ifdef HAVE_TK
-#    include "tkmain.h"
-#endif
+#include "wtools.h"		/* For common_dialog_repaint() */
 
 #define MAXLINKNAME 80
 #define HISTORY_SIZE 20
-#define HELP_WINDOW_WIDTH 62
+#define HELP_WINDOW_WIDTH (HELP_TEXT_WIDTH + 4)
 
-/* "$Id: help.c,v 1.10 2000/08/22 22:50:15 proskin Exp $" */
+#define STRING_LINK_START	"\01"
+#define STRING_LINK_POINTER	"\02"
+#define STRING_LINK_END		"\03"
+#define STRING_NODE_END		"\04"
 
 static char *data;		/* Pointer to the loaded data file */
-static int help_lines = 18;	/* Lines in help viewer */
+static int help_lines;		/* Lines in help viewer */
 static int  history_ptr;	/* For the history queue */
 static char *main_node;		/* The main node */
 static char *last_shown = 0;	/* Last byte shown in a screen */
 static int end_of_node = 0;	/* Flag: the last character of the node shown? */
-char *currentpoint, *startpoint;
+static char *currentpoint, *startpoint;
 static char *selected_item;
 
 /* The widget variables */
@@ -94,7 +93,7 @@ static int inside_link_area = 0;
 static int help_callback (struct Dlg_head *h, int id, int msg);
 
 #ifdef HAS_ACS_AS_PCCHARS
-struct {
+static const struct {
     int acscode;
     int pccode;
 } acs2pc_table [] = {
@@ -255,7 +254,7 @@ static void move_to_bottom (int dummy)
     move_backward (help_lines - 1);
 }
 
-char *help_follow_link (char *start, char *selected_item)
+static char *help_follow_link (char *start, char *selected_item)
 {
     char link_name [MAXLINKNAME];
     char *p;
@@ -276,7 +275,9 @@ char *help_follow_link (char *start, char *selected_item)
 	if (p)
 	    return p;
     }
-    return _(" Help file format error\n\x4");	/*  */
+
+    /* Create a replacement page with the error message */
+    return _(" Help file format error\n");
 }
 
 static char *select_next_link (char *start, char *current_link)
@@ -315,7 +316,7 @@ static void start_link_area (int x, int y, char *link_name)
     Link_Area *new;
 
     if (inside_link_area)
-	message (0, _(" Warning "), _(" Internal bug: Double start of link area "));
+	message (0, _("Warning"), _(" Internal bug: Double start of link area "));
 
     /* Allocate memory for a new link area */
     new = g_new (Link_Area, 1);
@@ -355,7 +356,7 @@ static void clear_link_areas (void)
     inside_link_area = 0;
 }
 
-static void show (Dlg_head *h, char *paint_start)
+static void help_show (Dlg_head *h, char *paint_start)
 {
     char *p;
     int  col, line, c;
@@ -364,6 +365,7 @@ static void show (Dlg_head *h, char *paint_start)
     int repeat_paint;
     int active_col, active_line;/* Active link position */
 
+    attrset (HELP_NORMAL_COLOR);
     do {
 	
 	line = col = acs = active_col = active_line = repeat_paint = 0;
@@ -372,8 +374,8 @@ static void show (Dlg_head *h, char *paint_start)
 	if (selected_item < paint_start)
 	    selected_item = NULL;
 	
-	for (p = paint_start; *p != CHAR_NODE_END && line < help_lines; p++){
-	    c = *p;
+	for (p = paint_start; *p && *p != CHAR_NODE_END && line < help_lines; p++) {
+	    c = (unsigned char)*p;
 	    switch (c){
 	    case CHAR_LINK_START:
 		if (selected_item == NULL)
@@ -408,13 +410,13 @@ static void show (Dlg_head *h, char *paint_start)
 		addstr (VERSION);
 		col += strlen (VERSION);
 		break;
-	    case CHAR_BOLD_ON:
+	    case CHAR_FONT_BOLD:
 		attrset (HELP_BOLD_COLOR);
 		break;
-	    case CHAR_ITALIC_ON:
+	    case CHAR_FONT_ITALIC:
 		attrset (HELP_ITALIC_COLOR);
 		break;
-	    case CHAR_BOLD_OFF:
+	    case CHAR_FONT_NORMAL:
 		attrset (HELP_NORMAL_COLOR);
 		break;
 	    case '\n':
@@ -424,16 +426,6 @@ static void show (Dlg_head *h, char *paint_start)
 	    case '\t':
 		col = (col/8 + 1) * 8;
 		break;
-	    case CHAR_MCLOGO:
-	    case CHAR_TEXTONLY_START:
-	    case CHAR_TEXTONLY_END:
-		break;
-	    case CHAR_XONLY_START:
-		while (*p && *p != CHAR_NODE_END && *p != CHAR_XONLY_END)
-		    p++;
-		if (*p == CHAR_NODE_END || !*p)
-		    p--;
-		break;
 	    default:
 		if (!painting)
 		    continue;
@@ -445,7 +437,7 @@ static void show (Dlg_head *h, char *paint_start)
 		    if (c == ' ' || c == '.')
 			addch (c);
 		    else
-#ifndef OS2_NT
+#ifndef NATIVE_WIN32
 #ifndef HAVE_SLANG
 			addch (acs_map [c]);
 #else
@@ -453,7 +445,7 @@ static void show (Dlg_head *h, char *paint_start)
 #endif
 #else
 			addch (acs2pc (c));
-#endif /* OS2_NT */
+#endif /* NATIVE_WIN32 */
 		} else
 		    addch (c);
 		col++;
@@ -559,24 +551,26 @@ help_help_cmd (Dlg_head *h)
 }
 
 static void
-help_index_cmd (Dlg_head *h)
+help_index_cmd (Dlg_head * h)
 {
     char *new_item;
 
-    history_ptr = (history_ptr+1) % HISTORY_SIZE;
-    history [history_ptr].page = currentpoint;
-    history [history_ptr].link = selected_item;
-    currentpoint = startpoint = search_string (data, "[Help]") + 1;
+    if (!(new_item = search_string (data, "[Contents]"))) {
+	message (1, MSG_ERROR, _(" Cannot find node %s in help file "),
+		 "[Contents]");
+	return;
+    }
 
-    if (!(new_item = search_string (data, "[Contents]")))
-	message (1, MSG_ERROR, _(" Can't find node [Contents] in help file "));
-    else
-	currentpoint = startpoint = new_item + 1;
+    history_ptr = (history_ptr + 1) % HISTORY_SIZE;
+    history[history_ptr].page = currentpoint;
+    history[history_ptr].link = selected_item;
+
+    currentpoint = startpoint = new_item + 1;
     selected_item = NULL;
     help_callback (h, 0, DLG_DRAW);
 }
 
-static void quit_cmd (void *x)
+static void help_quit_cmd (void *x)
 {
     dlg_stop ((Dlg_head *)x);
 }
@@ -592,9 +586,9 @@ static void prev_node_cmd (Dlg_head *h)
     help_callback (h, 0, DLG_DRAW);
 }
 
-static int md_callback (Dlg_head *h, Widget *w, int msg, int par)
+static int md_callback (Widget *w, int msg, int par)
 {
-    return default_proc (h, msg, par);
+    return default_proc (msg, par);
 }
 
 static Widget *mousedispatch_new (int y, int x, int yl, int xl)
@@ -691,10 +685,13 @@ static int help_handle_key (struct Dlg_head *h, int c)
 	while (*new_item && *new_item != CHAR_NODE_END)
 	    new_item++;
 	if (*++new_item == '['){
-	    while (*new_item != ']')
-		new_item++;
-	    currentpoint = new_item + 2;
-	    selected_item = NULL;
+	    while (*++new_item) {
+		if (*new_item == ']' && *++new_item && *++new_item) {
+		    currentpoint = new_item;
+		    selected_item = NULL;
+		    break;
+		}
+	    }
 	}
 	break;
 	
@@ -733,14 +730,8 @@ static int help_callback (struct Dlg_head *h, int id, int msg)
 {
     switch (msg){
     case DLG_DRAW:
-	attrset (HELP_NORMAL_COLOR);
-	dlg_erase (h);
-	draw_box (h, 1, 1, help_lines+2, HELP_WINDOW_WIDTH+2);
-	attrset (COLOR_HOT_NORMAL);
-	dlg_move (h, 1, (HELP_WINDOW_WIDTH - 1) / 2);
-	addstr (_(" Help "));
-	attrset (HELP_NORMAL_COLOR);
-	show (h, currentpoint);
+	common_dialog_repaint (h);
+	help_show (h, currentpoint);
 	break;
 
     case DLG_KEY:
@@ -760,68 +751,81 @@ void
 interactive_display (char *filename, char *node)
 {
     WButtonBar *help_bar;
-    Widget     *md;
-    
-    if ((data = load_file (filename)) == 0){
-	message (1, MSG_ERROR, _(" Can't open file %s \n %s "),
-		 filename, unix_error_string (errno));
-	return;
-    }
-    if (!(main_node = search_string (data, node))){
-	message (1, MSG_ERROR, _(" Can't find node %s in help file "), node);
-	interactive_display_finish ();
-	return;
+    Widget *md;
+    char *hlpfile = filename;
+
+    if (filename)
+	data = load_file (filename);
+    else
+	data = load_mc_home_file ("mc.hlp", &hlpfile);
+
+    if (data == NULL) {
+	message (1, MSG_ERROR, _(" Cannot open file %s \n %s "), hlpfile,
+		 unix_error_string (errno));
     }
 
-#ifndef HAVE_X
-    if (help_lines > LINES - 4)
-	help_lines = LINES - 4;
+    if (!filename)
+	g_free (hlpfile);
 
-    whelp = create_dlg (0, 0, help_lines+4, HELP_WINDOW_WIDTH+4, dialog_colors,
-			help_callback, "[Help]", "help", DLG_TRYUP|DLG_CENTER);
+    if (!data)
+	return;
 
-    /* allow us to process the tab key */
-    whelp->raw = 1;
+    if (!node || !*node)
+	node = "[main]";
+
+    if (!(main_node = search_string (data, node))) {
+	message (1, MSG_ERROR, _(" Cannot find node %s in help file "),
+		 node);
+
+	/* Fallback to [main], return if it also cannot be found */
+	main_node = search_string (data, "[main]");
+	if (!main_node) {
+	    interactive_display_finish ();
+	    return;
+	}
+    }
+
+    help_lines = min (LINES - 4, max (2 * LINES / 3, 18));
+
+    whelp =
+	create_dlg (0, 0, help_lines + 4, HELP_WINDOW_WIDTH + 4,
+		    dialog_colors, help_callback, "[Help]", _("Help"),
+		    DLG_TRYUP | DLG_CENTER | DLG_WANT_TAB);
 
-#endif    
     selected_item = search_string_node (main_node, STRING_LINK_START) - 1;
     currentpoint = startpoint = main_node + 1;
 
-    for (history_ptr = HISTORY_SIZE; history_ptr;){
+    for (history_ptr = HISTORY_SIZE; history_ptr;) {
 	history_ptr--;
-	history [history_ptr].page = currentpoint;
-	history [history_ptr].link = selected_item;
+	history[history_ptr].page = currentpoint;
+	history[history_ptr].link = selected_item;
     }
 
-#ifndef HAVE_X
-    help_bar = buttonbar_new (keybar_visible);
+    help_bar = buttonbar_new (1);
     help_bar->widget.y -= whelp->y;
     help_bar->widget.x -= whelp->x;
-    
-    md       = mousedispatch_new (1, 1, help_lines, HELP_WINDOW_WIDTH-2);
-    
+
+    md = mousedispatch_new (1, 1, help_lines, HELP_WINDOW_WIDTH - 2);
+
     add_widget (whelp, help_bar);
     add_widget (whelp, md);
 
-    define_label_data (whelp, (Widget *)NULL, 1, _("Help"),
-		       (buttonbarfn) help_help_cmd, whelp);
-    define_label_data (whelp, (Widget *)NULL, 2, _("Index"),
-		       (buttonbarfn) help_index_cmd,whelp);
-    define_label_data (whelp, (Widget *)NULL, 3, _("Prev"),
-		       (buttonbarfn) prev_node_cmd, whelp);
-    define_label (whelp, (Widget *) NULL, 4, "", 0);
-    define_label (whelp, (Widget *) NULL, 5, "", 0);
-    define_label (whelp, (Widget *) NULL, 6, "", 0);
-    define_label (whelp, (Widget *) NULL, 7, "", 0);
-    define_label (whelp, (Widget *) NULL, 8, "", 0);
-    define_label (whelp, (Widget *) NULL, 9, "", 0);
-    define_label_data (whelp, (Widget *) NULL, 10, _("Quit"), quit_cmd, whelp);
+    define_label_data (whelp, 1, _("Help"), (buttonbarfn) help_help_cmd,
+		       whelp);
+    define_label_data (whelp, 2, _("Index"), (buttonbarfn) help_index_cmd,
+		       whelp);
+    define_label_data (whelp, 3, _("Prev"), (buttonbarfn) prev_node_cmd,
+		       whelp);
+    define_label (whelp, 4, "", 0);
+    define_label (whelp, 5, "", 0);
+    define_label (whelp, 6, "", 0);
+    define_label (whelp, 7, "", 0);
+    define_label (whelp, 8, "", 0);
+    define_label (whelp, 9, "", 0);
+    define_label_data (whelp, 10, _("Quit"), help_quit_cmd, whelp);
 
     run_dlg (whelp);
     interactive_display_finish ();
     destroy_dlg (whelp);
-#else
-    x_interactive_display ();
-#endif
 }
 
