@@ -28,15 +28,11 @@
 #ifdef HAVE_UNISTD_H
 #   include <unistd.h>
 #endif
-#include <sys/types.h>
-#include <sys/time.h>
-#if HAVE_SYS_SELECT_H
-#   include <sys/select.h>
-#endif
-#include "tty.h"
 #include <ctype.h>
 #include <errno.h>
+
 #include "global.h"
+#include "tty.h"
 #include "mouse.h"
 #include "key.h"
 #include "main.h"
@@ -44,7 +40,7 @@
 #include "cons.saver.h"
 #include "../vfs/vfs.h"
 
-#if defined (HAVE_TEXTMODE_X11_SUPPORT) && !defined (HAVE_X)
+#ifdef HAVE_TEXTMODE_X11_SUPPORT
 #include <X11/Xlib.h>
 #endif
 
@@ -57,8 +53,6 @@
 #    include <sys/ioctl.h>
 #endif
 
-#include "x.h"
-
 #define GET_TIME(tv)    (gettimeofday(&tv, (struct timezone *)NULL))
 #define DIF_TIME(t1,t2) ((t2.tv_sec -t1.tv_sec) *1000+ \
 			 (t2.tv_usec-t1.tv_usec)/1000)
@@ -66,13 +60,19 @@
 /* timeout for old_esc_mode in usec */
 #define ESCMODE_TIMEOUT 1000000
 
+/* Linux console keyboard modifiers */
+#define SHIFT_PRESSED 1
+#define ALTR_PRESSED 2
+#define CONTROL_PRESSED 4
+#define ALTL_PRESSED 8
+
+
 int mou_auto_repeat = 100;
 int double_click_speed = 250;
 int old_esc_mode = 0;
 
 int use_8th_bit_as_meta = 1;
 
-#ifndef HAVE_X
 typedef struct key_def {
     char ch;			/* Holds the matching char code */
     int code;			/* The code returned, valid if child == NULL */
@@ -85,14 +85,12 @@ typedef struct key_def {
 
 /* This holds all the key definitions */
 static key_def *keys = 0;
-#endif
 
 static int input_fd;
-static fd_set select_set;
 static int disabled_channels = 0; /* Disable channels checking */
-int xgetch_second (void);
+static int xgetch_second (void);
+static int get_modifier (void);
 
-#ifndef PORT_HAS_FILE_HANDLERS
 /* File descriptor monitoring add/remove routines */
 typedef struct SelectList {
     int fd;
@@ -101,7 +99,7 @@ typedef struct SelectList {
     struct SelectList *next;
 } SelectList;
 
-SelectList *select_list = 0;
+static SelectList *select_list = 0;
 
 void add_select_channel (int fd, select_fn callback, void *info)
 {
@@ -118,17 +116,24 @@ void add_select_channel (int fd, select_fn callback, void *info)
 void delete_select_channel (int fd)
 {
     SelectList *p = select_list;
-    SelectList *prev = 0;
-    
-    while (p){
-	if (p->fd == fd){
-	    if (prev)
-		prev->next = p->next;
+    SelectList *p_prev = 0;
+    SelectList *p_next;
+
+    while (p) {
+	if (p->fd == fd) {
+	    p_next = p->next;
+
+	    if (p_prev)
+		p_prev->next = p_next;
 	    else
-		select_list = p->next;
+		select_list = p_next;
+
 	    g_free (p);
+	    p = p_next;
+	    continue;
 	}
-	prev = p;
+
+	p_prev = p;
 	p = p->next;
     }
 }
@@ -160,7 +165,6 @@ static void check_selects (fd_set *select_set)
 	if (FD_ISSET (p->fd, select_set))
 	    (*p->callback)(p->fd, p->info);
 }
-#endif
 
 void channels_down (void)
 {
@@ -175,26 +179,20 @@ void channels_up (void)
     disabled_channels--;
 }
 
-typedef struct {
+typedef const struct {
     int code;
     char *seq;
     int action;
 } key_define_t;
 
-#ifndef HAVE_X
-key_define_t mc_bindings [] = {
+static key_define_t mc_bindings [] = {
     { KEY_END,    ESC_STR ">", MCKEY_NOACTION },
     { KEY_HOME,   ESC_STR "<", MCKEY_NOACTION },
-
-#ifdef linux
-    /* Incredible, but many Linuxes still have old databases */
-    { KEY_IC,     ESC_STR "[2~", MCKEY_NOACTION },
-#endif
     { 0, 0, MCKEY_NOACTION },
 };
 
 /* Broken terminfo and termcap databases on xterminals */
-key_define_t xterm_key_defines [] = {
+static key_define_t xterm_key_defines [] = {
     { KEY_F(1),   ESC_STR "OP",   MCKEY_NOACTION },
     { KEY_F(2),   ESC_STR "OQ",   MCKEY_NOACTION },
     { KEY_F(3),   ESC_STR "OR",   MCKEY_NOACTION },
@@ -209,45 +207,54 @@ key_define_t xterm_key_defines [] = {
     { KEY_F(8),   ESC_STR "[19~", MCKEY_NOACTION },
     { KEY_F(9),   ESC_STR "[20~", MCKEY_NOACTION },
     { KEY_F(10),  ESC_STR "[21~", MCKEY_NOACTION },
+
+    /* xterm keys with modifiers */
+    { KEY_M_SHIFT | KEY_UP,    ESC_STR "O2A",   MCKEY_NOACTION },
+    { KEY_M_SHIFT | KEY_DOWN,  ESC_STR "O2B",   MCKEY_NOACTION },
+    { KEY_M_SHIFT | KEY_RIGHT, ESC_STR "O2C",   MCKEY_NOACTION },
+    { KEY_M_SHIFT | KEY_LEFT,  ESC_STR "O2D",   MCKEY_NOACTION },
+    { KEY_M_CTRL  | KEY_PPAGE, ESC_STR "[5;5~", MCKEY_NOACTION },
+    { KEY_M_CTRL  | KEY_NPAGE, ESC_STR "[6;5~", MCKEY_NOACTION },
+
+    /* rxvt keys with modifiers */
+    { KEY_M_SHIFT | KEY_UP,    ESC_STR "[a",  MCKEY_NOACTION },
+    { KEY_M_SHIFT | KEY_DOWN,  ESC_STR "[b",  MCKEY_NOACTION },
+    { KEY_M_SHIFT | KEY_RIGHT, ESC_STR "[c",  MCKEY_NOACTION },
+    { KEY_M_SHIFT | KEY_LEFT,  ESC_STR "[d",  MCKEY_NOACTION },
+    { KEY_M_CTRL  | KEY_PPAGE, ESC_STR "[5^", MCKEY_NOACTION },
+    { KEY_M_CTRL  | KEY_NPAGE, ESC_STR "[6^", MCKEY_NOACTION },
+    { KEY_M_CTRL  | KEY_HOME,  ESC_STR "[7^", MCKEY_NOACTION },
+    { KEY_M_CTRL  | KEY_END,   ESC_STR "[8^", MCKEY_NOACTION },
+    { KEY_M_SHIFT | KEY_HOME,  ESC_STR "[7$", MCKEY_NOACTION },
+    { KEY_M_SHIFT | KEY_END,   ESC_STR "[8$", MCKEY_NOACTION },
+
     { 0, 0, MCKEY_NOACTION },
 };
 
-key_define_t mc_default_keys [] = {
+static key_define_t mc_default_keys [] = {
     { ESC_CHAR,	ESC_STR, MCKEY_ESCAPE },
     { ESC_CHAR, ESC_STR ESC_STR, MCKEY_NOACTION },
     { 0, 0, MCKEY_NOACTION },
 };
-#endif
 
 static void
 define_sequences (key_define_t *kd)
 {
-#ifndef HAVE_X
     int i;
     
     for (i = 0; kd [i].code; i++)
 	define_sequence(kd [i].code, kd [i].seq, kd [i].action);
-#endif	
 }
 
-#if defined (HAVE_TEXTMODE_X11_SUPPORT) && !defined (HAVE_X)
-Display *display;
-Window w;
-
-void
-init_textmode_x11_support (void)
-{
-    display = XOpenDisplay (0);
-    if (display)
-        w = DefaultRootWindow (display);
-}
-#endif
+#ifdef HAVE_TEXTMODE_X11_SUPPORT
+static Display *x11_display;
+static Window x11_window;
+#endif /* HAVE_TEXTMODE_X11_SUPPORT */
 
 /* This has to be called before slang_init or whatever routine
    calls any define_sequence */
 void init_key (void)
 {
-#ifndef HAVE_X
     char *term = (char *) getenv ("TERM");
     
     /* This has to be the first define_sequence */
@@ -281,21 +288,23 @@ void init_key (void)
 	use_8th_bit_as_meta = 0;
     }
 #endif /* __QNX__ */
-#endif /* !HAVE_X */
+
+#ifdef HAVE_TEXTMODE_X11_SUPPORT
+    x11_display = XOpenDisplay (0);
+    if (x11_display)
+        x11_window = DefaultRootWindow (x11_display);
+#endif /* HAVE_TEXTMODE_X11_SUPPORT */
 }
 
 /* This has to be called after SLang_init_tty/slint_init */
 void init_key_input_fd (void)
 {
-#ifndef HAVE_X
 #ifdef HAVE_SLANG
     input_fd = SLang_TT_Read_FD;
 #endif
-#endif /* !HAVE_X */
 }
 
 
-#ifndef HAVE_X
 static void
 xmouse_get_event (Gpm_Event *ev)
 {
@@ -303,21 +312,28 @@ xmouse_get_event (Gpm_Event *ev)
     static struct timeval tv1 = { 0, 0 }; /* Force first click as single */
     static struct timeval tv2;
     static int clicks;
+    static int last_btn = 0;
 
     /* Decode Xterm mouse information to a GPM style event */
 
     /* Variable btn has following meaning: */
     /* 0 = btn1 dn, 1 = btn2 dn, 2 = btn3 dn, 3 = btn up */
-    btn = xgetch () - 32;
+    btn = getch () - 32;
     
     /* There seems to be no way of knowing which button was released */
     /* So we assume all the buttons were released */
 
     if (btn == 3){
-        ev->type = GPM_UP | (GPM_SINGLE << clicks);
-        ev->buttons = 0;
-	GET_TIME (tv1);
-	clicks = 0;
+	if (last_btn) {
+	    ev->type = GPM_UP | (GPM_SINGLE << clicks);
+	    ev->buttons = 0;
+	    last_btn = 0;
+	    GET_TIME (tv1);
+	    clicks = 0;
+	} else {
+	    /* Bogus event, maybe mouse wheel */
+	    ev->type = 0;
+	}
     } else {
         ev->type = GPM_DOWN;
 	GET_TIME (tv2);
@@ -337,15 +353,24 @@ xmouse_get_event (Gpm_Event *ev)
 	case 2:
             ev->buttons = GPM_B_RIGHT;
             break;
+	case 64:
+            ev->buttons = GPM_B_UP;
+            break;
+	case 65:
+            ev->buttons = GPM_B_DOWN;
+            break;
 	default:
             /* Nothing */
+	    ev->type = 0;
+	    ev->buttons = 0;
             break;
         }
+	last_btn = ev->buttons;
     }
     /* Coordinates are 33-based */
     /* Transform them to 1-based */
-    ev->x = xgetch () - 32;
-    ev->y = xgetch () - 32;
+    ev->x = getch () - 32;
+    ev->y = getch () - 32;
 }
 
 static key_def *create_sequence (char *seq, int code, int action)
@@ -370,7 +395,8 @@ static key_def *create_sequence (char *seq, int code, int action)
 }
 
 /* The maximum sequence length (32 + null terminator) */
-static int seq_buffer [33];
+#define SEQ_BUFFER_LEN 33
+static int seq_buffer [SEQ_BUFFER_LEN];
 static int *seq_append = 0;
 
 static int push_char (int c)
@@ -378,33 +404,35 @@ static int push_char (int c)
     if (!seq_append)
 	seq_append = seq_buffer;
     
-    if (seq_append == &(seq_buffer [sizeof (seq_buffer)-2]))
+    if (seq_append == &(seq_buffer [SEQ_BUFFER_LEN-2]))
 	return 0;
     *(seq_append++) = c;
     *seq_append = 0;
     return 1;
 }
-#endif /* !HAVE_X */
 
-void define_sequence (int code, char *seq, int action)
+/*
+ * Return 1 on success, 0 on error.
+ * An error happens if SEQ is a beginning of an existing longer sequence.
+ */
+int define_sequence (int code, char *seq, int action)
 {
-#ifndef HAVE_X
     key_def *base;
 
-    if (strlen (seq) > sizeof (seq_buffer)-1)
-	return;
+    if (strlen (seq) > SEQ_BUFFER_LEN-1)
+	return 0;
     
     for (base = keys; (base != 0) && *seq; ){
 	if (*seq == base->ch){
 	    if (base->child == 0){
 		if (*(seq+1)){
 		    base->child = create_sequence (seq+1, code, action);
-		    return;
+		    return 1;
 		} else {
-		    /* The sequence clashes */
+		    /* The sequence matches an existing one.  */
                     base->code = code;
                     base->action = action;
-		    return;
+		    return 1;
 		}
 	    } else {
 		base = base->child;
@@ -415,47 +443,93 @@ void define_sequence (int code, char *seq, int action)
 		base = base->next;
 	    else {
 		base->next = create_sequence (seq, code, action);
-		return;
+		return 1;
 	    }
 	}
     }
+
+    if (!*seq) {
+	/* Attempt to redefine a sequence with a shorter sequence.  */
+	return 0;
+    }
+
     keys = create_sequence (seq, code, action);
-#endif    
+    return 1;
 }
 
-#ifndef HAVE_X
 static int *pending_keys;
-#endif
 
+/* Apply corrections for the keycode generated in get_key_code() */
 static int
-correct_key_code (int c)
+correct_key_code (int code)
 {
-    /* This is needed on some OS that do not support ncurses and */
-    /* do some magic after read()ing the data */
+    unsigned int c = code & ~KEY_M_MASK;	/* code without modifier */
+    unsigned int mod = code & KEY_M_MASK;	/* modifier */
+
+    /*
+     * Add key modifiers directly from X11 or OS.
+     * Ordinary characters only get modifiers from sequences.
+     */
+    if (c < 32 || c >= 256) {
+	mod |= get_modifier ();
+    }
+
+    /* This is needed if the newline is reported as carriage return */
     if (c == '\r')
-	return '\n';
+	c = '\n';
 
-#ifdef IS_AIX
+    /* This is reported to be useful on AIX */
     if (c == KEY_SCANCEL)
-	return '\t';
-#endif
+	c = '\t';
 
-    if (c == KEY_F(0))
-	return KEY_F(10);
+    /* Convert Shift+Tab and Ctrl+Tab to Back Tab */
+    if ((c == '\t') && (mod & (KEY_M_SHIFT | KEY_M_CTRL))) {
+	c = KEY_BTAB;
+	mod = 0;
+    }
 
-    if (!alternate_plus_minus) 
-        switch (c) {
-            case KEY_KP_ADD: c = '+'; break;
-            case KEY_KP_SUBTRACT: c = '-'; break;
-            case KEY_KP_MULTIPLY: c = '*'; break;
-        }
+    /* F0 is the same as F10 for out purposes */
+    if (c == KEY_F (0))
+	c = KEY_F (10);
 
-    return c;
+    /*
+     * We are not interested if Ctrl was pressed when entering control
+     * characters, so assume that it was.  When checking for such keys,
+     * XCTRL macro should be used.  In some cases, we are interested,
+     * e.g. to distinguish Ctrl-Enter from Enter.
+     */
+    if (c < 32 && c != ESC_CHAR && c != '\t' && c != '\n') {
+	mod |= KEY_M_CTRL;
+    }
+
+    /* Convert Shift+Fn to F(n+10) */
+    if (c >= KEY_F (1) && c <= KEY_F (10) && (mod & KEY_M_SHIFT)) {
+	c += 10;
+    }
+
+    /* Remove Shift information from function keys */
+    if (c >= KEY_F (1) && c <= KEY_F (20)) {
+	mod &= ~KEY_M_SHIFT;
+    }
+
+    if (!alternate_plus_minus)
+	switch (c) {
+	case KEY_KP_ADD:
+	    c = '+';
+	    break;
+	case KEY_KP_SUBTRACT:
+	    c = '-';
+	    break;
+	case KEY_KP_MULTIPLY:
+	    c = '*';
+	    break;
+	}
+
+    return (mod | c);
 }
 
 int get_key_code (int no_delay)
 {
-#ifndef HAVE_X
     int c;
     static key_def *this = NULL, *parent;
     static struct timeval esctime = { -1, -1 };
@@ -486,26 +560,22 @@ int get_key_code (int no_delay)
 
  nodelay_try_again:
     if (no_delay) {
-#ifdef BUGGY_CURSES
-        wtimeout(stdscr, 500);
-#else
         nodelay (stdscr, TRUE);
-#endif
     }
-    c = xgetch ();
-    if (no_delay) {
-#ifdef BUGGY_CURSES
-        notimeout (stdscr, TRUE);
-#else
-        nodelay (stdscr, FALSE);
+    c = getch ();
+#if defined(USE_NCURSES) && defined(KEY_RESIZE)
+    if (c == KEY_RESIZE)
+        goto nodelay_try_again;
 #endif
-        if (c == ERR) {
+    if (no_delay) {
+        nodelay (stdscr, FALSE);
+        if (c == -1) {
             if (this != NULL && parent != NULL && 
                 parent->action == MCKEY_ESCAPE && old_esc_mode) {
                 struct timeval current, timeout;
                 
                 if (esctime.tv_sec == -1)
-                    return ERR;
+                    return -1;
                 GET_TIME (current);
                 timeout.tv_sec = ESCMODE_TIMEOUT / 1000000 + esctime.tv_sec;
                 timeout.tv_usec = ESCMODE_TIMEOUT % 1000000 + esctime.tv_usec;
@@ -514,26 +584,26 @@ int get_key_code (int no_delay)
                     timeout.tv_sec++;
                 }
                 if (current.tv_sec < timeout.tv_sec)
-                    return ERR;
+                    return -1;
                 if (current.tv_sec == timeout.tv_sec && 
                     current.tv_usec < timeout.tv_usec)
-                    return ERR;
+                    return -1;
                 this = NULL;
 		pending_keys = seq_append = NULL;
 		return ESC_CHAR;
             }
-            return ERR;
+            return -1;
         }
-    } else if (c == ERR){
+    } else if (c == -1){
 	/* Maybe we got an incomplete match.
 	   This we do only in delay mode, since otherwise
-	   xgetch can return ERR at any time. */
+	   getch can return -1 at any time. */
 	if (seq_append) {
 	    pending_keys = seq_buffer;
 	    goto pend_send;
 	}
 	this = NULL;
-	return ERR;
+	return -1;
     }
     
     /* Search the key on the root */
@@ -570,7 +640,7 @@ int get_key_code (int no_delay)
 		    }
 		    esctime.tv_sec = -1;
 		    c = xgetch_second ();
-		    if (c == ERR) {
+		    if (c == -1) {
 		        pending_keys = seq_append = NULL;
 		        this = NULL;
 		        return ESC_CHAR;
@@ -578,7 +648,7 @@ int get_key_code (int no_delay)
 		} else {
 		    if (no_delay)
 		        goto nodelay_try_again;
-		    c = xgetch ();
+		    c = getch ();
 		}
 	    } else {
 		/* We got a complete match, return and reset search */
@@ -620,12 +690,8 @@ int get_key_code (int no_delay)
     }
     this = NULL;
     return correct_key_code (c);
-#else
-    return ERR;
-#endif /* HAVE_X */
 }
 
-#ifndef PORT_HAS_FILE_HANDLERS
 /* If set timeout is set, then we wait 0.1 seconds, else, we block */
 static void
 try_channels (int set_timeout)
@@ -634,11 +700,12 @@ try_channels (int set_timeout)
     static fd_set select_set;
     struct timeval *timeptr;
     int v;
+    int maxfdp;
 
     while (1){
 	FD_ZERO (&select_set);
 	FD_SET  (input_fd, &select_set);	/* Add stdin */
-	add_selects (&select_set);
+	maxfdp = max (add_selects (&select_set), input_fd);
 	
 	if (set_timeout){
 	    timeout.tv_sec = 0;
@@ -647,7 +714,7 @@ try_channels (int set_timeout)
 	} else
 	    timeptr = 0;
 
-	v = select (FD_SETSIZE, &select_set, NULL, NULL, timeptr);
+	v = select (maxfdp + 1, &select_set, NULL, NULL, timeptr);
 	if (v > 0){
 	    check_selects (&select_set);
 	    if (FD_ISSET (input_fd, &select_set))
@@ -656,7 +723,6 @@ try_channels (int set_timeout)
     }
 }
 
-#ifndef HAVE_X
 /* Workaround for System V Curses vt100 bug */
 static int getch_with_delay (void)
 {
@@ -670,7 +736,7 @@ static int getch_with_delay (void)
 
 	/* Try to get a character */
 	c = get_key_code (0);
-	if (c != ERR)
+	if (c != -1)
 	    break;
 	/* Failed -> wait 0.1 secs and try again */
 	try_channels (1);
@@ -678,32 +744,24 @@ static int getch_with_delay (void)
     /* Success -> return the character */
     return c;
 }
-#endif /* !HAVE_X */
 
-#ifndef HAVE_LIBGPM
-#define gpm_flag 0
-#endif
-#endif /* !HAVE_FILE_HANDLERS */
-
-extern int max_dirt_limit;
-
-#ifndef HAVE_X
 /* Returns a character read from stdin with appropriate interpretation */
 /* Also takes care of generated mouse events */
 /* Returns EV_MOUSE if it is a mouse event */
 /* Returns EV_NONE  if non-blocking or interrupt set and nothing was done */
-int get_event (Gpm_Event *event, int redo_event, int block)
+int
+get_event (Gpm_Event * event, int redo_event, int block)
 {
     int c;
-    static int flag;			/* Return value from select */
+    static int flag;		/* Return value from select */
 #ifdef HAVE_LIBGPM
-    static Gpm_Event ev;		/* Mouse event */
+    static Gpm_Event ev;	/* Mouse event */
 #endif
     struct timeval timeout;
     struct timeval *time_addr = NULL;
     static int dirty = 3;
 
-    if ((dirty == 3) || is_idle ()){
+    if ((dirty == 3) || is_idle ()) {
 	mc_refresh ();
 	doupdate ();
 	dirty = 1;
@@ -711,120 +769,124 @@ int get_event (Gpm_Event *event, int redo_event, int block)
 	dirty++;
 
     vfs_timeout_handler ();
-    
+
     /* Ok, we use (event->x < 0) to signal that the event does not contain
        a suitable position for the mouse, so we can't use show_mouse_pointer
        on it.
-    */
-    if (event->x > 0){
+     */
+    if (event->x > 0) {
 	show_mouse_pointer (event->x, event->y);
 	if (!redo_event)
 	    event->x = -1;
     }
 
     /* Repeat if using mouse */
-    while ((xmouse_flag || gpm_flag) && !pending_keys)
-    {
-	if (xmouse_flag || gpm_flag)
-	{
-	    FD_ZERO (&select_set);
-	    FD_SET  (input_fd, &select_set);
-	    add_selects (&select_set);
+    while (mouse_enabled && !pending_keys) {
+	int maxfdp;
+	fd_set select_set;
+
+	FD_ZERO (&select_set);
+	FD_SET (input_fd, &select_set);
+	maxfdp = max (add_selects (&select_set), input_fd);
 
 #ifdef HAVE_LIBGPM
-	    if (gpm_flag) {
-		FD_SET  (gpm_fd, &select_set);
+	if (use_mouse_p == MOUSE_GPM) {
+	    if (gpm_fd == -1) {
+		/* Connection to gpm broken, possibly gpm has died */
+		mouse_enabled = 0;
+		use_mouse_p = MOUSE_NONE;
+		break;
 	    }
+	    FD_SET (gpm_fd, &select_set);
+	    maxfdp = max (maxfdp, gpm_fd);
+	}
 #endif
 
-	    if (redo_event){
-	        timeout.tv_usec = mou_auto_repeat * 1000;
-		timeout.tv_sec = 0;
+	if (redo_event) {
+	    timeout.tv_usec = mou_auto_repeat * 1000;
+	    timeout.tv_sec = 0;
 
-		time_addr = &timeout;
-	    } else {
-		int seconds;
-		
-		if ((seconds = vfs_timeouts ())){
-		    /* the timeout could be improved and actually be
-		     * the number of seconds until the next vfs entry
-		     * timeouts in the stamp list.
-		     */
+	    time_addr = &timeout;
+	} else {
+	    int seconds;
 
-		    timeout.tv_sec = seconds;
-		    timeout.tv_usec = 0;
-		    time_addr = &timeout;
-		} else
-		    time_addr = NULL;
-	    }
+	    if ((seconds = vfs_timeouts ())) {
+		/* the timeout could be improved and actually be
+		 * the number of seconds until the next vfs entry
+		 * timeouts in the stamp list.
+		 */
 
-	    if (!block){
-		time_addr = &timeout;
-		timeout.tv_sec = 0;
+		timeout.tv_sec = seconds;
 		timeout.tv_usec = 0;
-	    }
-	    enable_interrupt_key ();
-	    flag = select (FD_SETSIZE, &select_set, NULL, NULL, time_addr);
-	    disable_interrupt_key ();
-	    
-	    /* select timed out: it could be for any of the following reasons:
-	     * redo_event -> it was because of the MOU_REPEAT handler
-	     * !block     -> we did not block in the select call
-	     * else       -> 10 second timeout to check the vfs status.
-	     */
-	    if (flag == 0){
-		if (redo_event)
-		    return EV_MOUSE;
-		if (!block)
-		    return EV_NONE;
-		vfs_timeout_handler ();
-	    }
-	    if (flag == -1 && errno == EINTR)
-		return EV_NONE;
-
-	    check_selects (&select_set);
-	    
-	    if (FD_ISSET (input_fd, &select_set))
-	        break;
+		time_addr = &timeout;
+	    } else
+		time_addr = NULL;
 	}
+
+	if (!block) {
+	    time_addr = &timeout;
+	    timeout.tv_sec = 0;
+	    timeout.tv_usec = 0;
+	}
+	enable_interrupt_key ();
+	flag = select (maxfdp + 1, &select_set, NULL, NULL, time_addr);
+	disable_interrupt_key ();
+
+	/* select timed out: it could be for any of the following reasons:
+	 * redo_event -> it was because of the MOU_REPEAT handler
+	 * !block     -> we did not block in the select call
+	 * else       -> 10 second timeout to check the vfs status.
+	 */
+	if (flag == 0) {
+	    if (redo_event)
+		return EV_MOUSE;
+	    if (!block)
+		return EV_NONE;
+	    vfs_timeout_handler ();
+	}
+	if (flag == -1 && errno == EINTR)
+	    return EV_NONE;
+
+	check_selects (&select_set);
+
+	if (FD_ISSET (input_fd, &select_set))
+	    break;
 #ifdef HAVE_LIBGPM
-	if (gpm_flag && FD_ISSET (gpm_fd, &select_set)){
-	    if (gpm_flag){
-		Gpm_GetEvent (&ev);
-		Gpm_FitEvent (&ev);
-	    }
+	if (use_mouse_p == MOUSE_GPM && FD_ISSET (gpm_fd, &select_set)) {
+	    Gpm_GetEvent (&ev);
+	    Gpm_FitEvent (&ev);
 	    *event = ev;
 	    return EV_MOUSE;
 	}
-#endif
+#endif				/* !HAVE_LIBGPM */
     }
-#   ifndef HAVE_SLANG
-    flag = is_wintouched(stdscr);
+#ifndef HAVE_SLANG
+    flag = is_wintouched (stdscr);
     untouchwin (stdscr);
-#   endif
-    c = block ? getch_with_delay () : get_key_code(1);
+#endif				/* !HAVE_SLANG */
+    c = block ? getch_with_delay () : get_key_code (1);
 
-#   ifndef HAVE_SLANG
+#ifndef HAVE_SLANG
     if (flag)
-        touchwin (stdscr);
-#   endif
-    
-    if (c == MCKEY_MOUSE) { /* Mouse event */
-        xmouse_get_event (event);
-        return EV_MOUSE;
+	touchwin (stdscr);
+#endif				/* !HAVE_SLANG */
+
+    if (c == MCKEY_MOUSE
+#ifdef KEY_MOUSE
+	|| c == KEY_MOUSE
+#endif				/* KEY_MOUSE */
+	) {
+	/* Mouse event */
+	xmouse_get_event (event);
+	if (event->type)
+	    return EV_MOUSE;
+	else
+	    return EV_NONE;
     }
 
     return c;
 }
-#else
-/* X11 version of the get_event routine */
-int get_event (Gpm_Event *event, int redo_event, int block)
-{
-    return EV_NONE;
-}
-#endif
 
-#ifndef PORT_HAS_GETCH
 /* Returns a key press, mouse events are discarded */
 int mi_getch ()
 {
@@ -837,7 +899,7 @@ int mi_getch ()
     return key;
 }
 
-int xgetch_second (void)
+static int xgetch_second (void)
 {
     fd_set Read_FD_Set;
     int c;
@@ -845,25 +907,15 @@ int xgetch_second (void)
 
     timeout.tv_sec = ESCMODE_TIMEOUT / 1000000;
     timeout.tv_usec = ESCMODE_TIMEOUT % 1000000;
-#ifdef BUGGY_CURSES
-    wtimeout(stdscr, 500);
-#else
     nodelay (stdscr, TRUE);
-#endif
     FD_ZERO (&Read_FD_Set);
     FD_SET (input_fd, &Read_FD_Set);
-    select (FD_SETSIZE, &Read_FD_Set, NULL, NULL, &timeout);
-    c = xgetch ();
-#ifdef BUGGY_CURSES
-    notimeout (stdscr, TRUE);
-#else
+    select (input_fd + 1, &Read_FD_Set, NULL, NULL, &timeout);
+    c = getch ();
     nodelay (stdscr, FALSE);
-#endif
     return c;
 }
-#endif
 
-#ifndef HAVE_X
 static void
 learn_store_key (char *buffer, char **p, int c)
 {
@@ -890,12 +942,14 @@ char *learn_key (void)
     fd_set Read_FD_Set;
     struct timeval endtime;
     struct timeval timeout;
-    int c = xgetch ();
+    int c;
     char buffer [256];
     char *p = buffer;
     
-    while (c == ERR)
-        c = xgetch (); /* Sanity check, should be unnecessary */
+    keypad(stdscr, FALSE); /* disable intepreting keys by ncurses */
+    c = getch ();
+    while (c == -1)
+        c = getch (); /* Sanity check, should be unnecessary */
     learn_store_key (buffer, &p, c);
     GET_TIME (endtime);
     endtime.tv_usec += LEARN_TIMEOUT;
@@ -903,13 +957,9 @@ char *learn_key (void)
         endtime.tv_usec -= 1000000;
         endtime.tv_sec++;
     }
-#ifdef BUGGY_CURSES
-    wtimeout(stdscr, 500);
-#else
     nodelay (stdscr, TRUE);
-#endif
     for (;;) {
-        while ((c = xgetch ()) == ERR) {
+        while ((c = getch ()) == -1) {
             GET_TIME (timeout);
             timeout.tv_usec = endtime.tv_usec - timeout.tv_usec;
             if (timeout.tv_usec < 0)
@@ -918,19 +968,16 @@ char *learn_key (void)
             if (timeout.tv_sec >= 0 && timeout.tv_usec > 0) {
     		FD_ZERO (&Read_FD_Set);
     		FD_SET (input_fd, &Read_FD_Set);
-                select (FD_SETSIZE, &Read_FD_Set, NULL, NULL, &timeout);
+                select (input_fd + 1, &Read_FD_Set, NULL, NULL, &timeout);
             } else
             	break;
         }
-        if (c == ERR)
+        if (c == -1)
             break;
-	learn_store_key (buffer, &p, c);        
+	learn_store_key (buffer, &p, c);
     }
-#ifdef BUGGY_CURSES
-    notimeout (stdscr, TRUE);
-#else
+    keypad(stdscr, TRUE);
     nodelay (stdscr, FALSE);
-#endif
     *p = 0;
     return g_strdup (buffer);
 }
@@ -956,81 +1003,85 @@ application_keypad_mode (void)
     }
 }
 
-#endif /* !HAVE_X */
 
-/* A function to check if we're idle.
-   Currently checks only for key presses.
-   We could also check the mouse. */
-int is_idle (void)
+/*
+ * Check if we are idle, i.e. there are no pending keyboard or mouse
+ * events.  Return 1 is idle, 0 is there are pending events.
+ */
+int
+is_idle (void)
 {
-    /* Check for incoming key presses     *
-     * If there are any we say we're busy */
-
+    int maxfdp;
     fd_set select_set;
     struct timeval timeout;
+
     FD_ZERO (&select_set);
-    FD_SET (0, &select_set);
+    FD_SET (input_fd, &select_set);
+    maxfdp = input_fd;
+#ifdef HAVE_LIBGPM
+    if (use_mouse_p == MOUSE_GPM && mouse_enabled && gpm_fd != -1) {
+	FD_SET (gpm_fd, &select_set);
+	maxfdp = max (maxfdp, gpm_fd);
+    }
+#endif
     timeout.tv_sec = 0;
     timeout.tv_usec = 0;
-    select (FD_SETSIZE, &select_set, 0, 0, &timeout);
-    return ! FD_ISSET (0, &select_set);
+    return (select (maxfdp + 1, &select_set, 0, 0, &timeout) <= 0);
 }
 
 
-int
+/*
+ * Get modifier state (shift, alt, ctrl) for the last key pressed.
+ * We are assuming that the state didn't change since the key press.
+ * This is only correct if get_modifier() is called very fast after
+ * the input was received, so that the user didn't release the
+ * modifier keys yet.
+ */
+static int
 get_modifier (void)
 {
-#if defined (HAVE_TEXTMODE_X11_SUPPORT) && !defined (HAVE_X)
-    if (display) {
-        Window root, child;
-        int root_x, root_y;
-        int win_x, win_y;
-        unsigned int mask;
-        Bool b;
-	int result = 0;
+    int result = 0;
 
-	b = XQueryPointer(display, w, &root, &child,
-                                  &root_x, &root_y,
-                                  &win_x, &win_y,
-             			  &mask);
+#ifdef HAVE_TEXTMODE_X11_SUPPORT
+    if (x11_display) {
+	Window root, child;
+	int root_x, root_y;
+	int win_x, win_y;
+	unsigned int mask;
+
+	XQueryPointer (x11_display, x11_window, &root, &child, &root_x,
+		       &root_y, &win_x, &win_y, &mask);
 
 	if (mask & ShiftMask)
-	    result |= SHIFT_PRESSED;
-        if (mask & ControlMask)
-	    result |= CONTROL_PRESSED;
+	    result |= KEY_M_SHIFT;
+	if (mask & ControlMask)
+	    result |= KEY_M_CTRL;
 	return result;
-    } else 
+    }
 #endif
 #ifdef __linux__
     {
-	unsigned char modifiers;
-
-	modifiers = 6;
+	unsigned char modifiers = 6;
 
 	if (ioctl (0, TIOCLINUX, &modifiers) < 0)
 	    return 0;
 
-        return (int) modifiers;
+	/* Translate Linux modifiers into mc modifiers */
+	if (modifiers & SHIFT_PRESSED)
+	    result |= KEY_M_SHIFT;
+	if (modifiers & (ALTL_PRESSED | ALTR_PRESSED))
+	    result |= KEY_M_ALT;
+	if (modifiers & CONTROL_PRESSED)
+	    result |= KEY_M_CTRL;
+
+	return result;
     }
 #else
-    return 0;
+    return result;
 #endif
 }
 
-int
-ctrl_pressed ()
-{
-#ifdef __linux__
-    if (get_modifier () & CONTROL_PRESSED)
-	return 1;
-#endif
-    return 0;
-}
-
-#ifdef HAVE_MAD
-#ifndef HAVE_X
-
-void k_dispose (key_def *k)
+static void k_dispose (key_def *k)
 {
     if (!k)
 	return;
@@ -1039,7 +1090,7 @@ void k_dispose (key_def *k)
     g_free (k);
 }
 
-void s_dispose (SelectList *sel)
+static void s_dispose (SelectList *sel)
 {
     if (!sel)
 	return;
@@ -1052,21 +1103,9 @@ void done_key ()
 {
     k_dispose (keys);
     s_dispose (select_list);
-}
 
-#else
-void done_key () 
-{
+#ifdef HAVE_TEXTMODE_X11_SUPPORT
+    if (x11_display)
+	XCloseDisplay (x11_display);
+#endif /* HAVE_TEXTMODE_X11_SUPPORT */
 }
-
-#endif /* HAVE_X */
-#endif /* HAVE_MAD */
-
-#if defined (HAVE_TEXTMODE_X11_SUPPORT) && !defined (HAVE_X)
-void
-done_textmode_x11_support (void)
-{
-    if (display)
-	XCloseDisplay (display);
-}
-#endif
