@@ -23,102 +23,73 @@
 
 #include <config.h>
 #include <stdio.h>
-#if defined(NEEDS_IO_H)            /* OS/2 need io.h! .ado */
-#    include <io.h>
-#endif
 #include <sys/types.h>
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+#   include <unistd.h>
 #endif
-#include <fcntl.h>
 #include <signal.h>		/* my_system */
 #include <limits.h>		/* INT_MAX */
-#ifndef SCO_FLAVOR
-#	include <sys/time.h>	/* alex: sys/select.h defines struct timeval */
-#endif /* SCO_FLAVOR */
-#include <sys/param.h>
-#include <sys/types.h>
 #include <sys/stat.h>
 #include <stdarg.h>
 #include <errno.h>		/* my_system */
-#ifdef SCO_FLAVOR
-#	include <sys/timeb.h>	/* alex: for struct timeb, used in time.h */
-#endif /* SCO_FLAVOR */
-#include <time.h>
-#ifndef HAS_NO_GRP_PWD_H
-#   include <pwd.h>
-#   include <grp.h>
-#endif
 #include <string.h>
 #include <ctype.h>
-#ifdef HAVE_SYS_SELECT_H
-#  include <sys/select.h>
-#endif
 
-#ifdef __linux__
-#    if defined(__GLIBC__) && (__GLIBC__ < 2)
-#        include <linux/termios.h>	/* This is needed for TIOCLINUX */
-#    else
-#        include <termios.h>
-#    endif
-#  include <sys/ioctl.h>
-#endif
-
-#include "mountlist.h"
-
-#if defined(HAVE_RX_H) && defined(HAVE_REGCOMP)
-#include <rx.h>
-#else
-#include <regex.h>
-#endif
 #include "global.h"
 #include "profile.h"
-#include "user.h"		/* expand_format */
+#include "main.h"		/* mc_home */
+#include "cmd.h"		/* guess_message_value */
 #include "../vfs/vfs.h"
+#include "mountlist.h"
+#include "win.h"		/* xterm_flag */
 
-/* "$Id: util.c,v 1.33 2000/12/18 16:42:20 alan Exp $" */
-
-char app_text [] = "Midnight-Commander";
-
-int easy_patterns = 1;
-int align_extensions = 1;
-int tilde_trunc = 1;
-
-struct mount_entry *mount_list = NULL;
-
-#ifndef VFS_STANDALONE
-int is_printable (int c)
-{
-    static const unsigned char xterm_printable[] = {
-        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
-        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
-        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
-        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,
-        1,1,1,1,0,0,1,1,0,1,1,1,1,0,0,0,0,1,1,1,1,1,0,0,0,1,0,0,0,0,0,0,
-        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
-        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
-        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
-    };
-
-    extern int xterm_flag;
-    extern int eight_bit_clean;
-    extern int full_eight_bits;
-
-#ifdef HAVE_GNOME
-    return 1;
+#ifdef HAVE_CHARSET
+#include "charsets.h"
 #endif
-    
+
+static const char app_text [] = "Midnight-Commander";
+int easy_patterns = 1;
+
+static inline int
+is_7bit_printable (unsigned char c)
+{
+    return (c > 31 && c < 127);
+}
+
+static inline int
+is_iso_printable (unsigned char c)
+{
+    return ((c > 31 && c < 127) || c >= 160);
+}
+
+static inline int
+is_8bit_printable (unsigned char c)
+{
+    /* "Full 8 bits output" doesn't work on xterm */
+    if (xterm_flag)
+	return is_iso_printable (c);
+
+    return (c > 31 && c != 127 && c != 155);
+}
+
+int
+is_printable (int c)
+{
     c &= 0xff;
-    if (eight_bit_clean){
-        if (full_eight_bits){
-	    if (xterm_flag)
-	        return xterm_printable [c];
-	    else
-	        return (c > 31 && c != 127 && c != 155);
-	} else
-	    return ((c >31 && c < 127) || c >= 160);
+
+#ifdef HAVE_CHARSET
+    /* "Display bits" is ignored, since the user controls the output
+       by setting the output codepage */
+    return is_8bit_printable (c);
+#else
+    if (!eight_bit_clean)
+	return is_7bit_printable (c);
+
+    if (full_eight_bits) {
+	return is_8bit_printable (c);
     } else
-	return (c > 31 && c < 127);
+	return is_iso_printable (c);
+#endif				/* !HAVE_CHARSET */
 }
 
 /* Returns the message dimensions (lines and columns) */
@@ -140,68 +111,90 @@ int msglen (char *text, int *lines)
     return max;
 }
 
-char *trim (char *s, char *d, int len)
+/*
+ * Copy from s to d, and trim the beginning if necessary, and prepend
+ * "..." in this case.  The destination string can have at most len
+ * bytes, not counting trailing 0.
+ */
+char *
+trim (char *s, char *d, int len)
 {
-    int source_len = strlen (s);
-    
-    if (source_len > len){
-	strcpy (d, s+(source_len-len));
-	d [0] = '.';
-	d [1] = '.';
-	d [2] = '.';
+    int source_len;
+
+    /* Sanity check */
+    len = max (len, 0);
+
+    source_len = strlen (s);
+    if (source_len > len) {
+	/* Cannot fit the whole line */
+	if (len <= 3) {
+	    /* We only have room for the dots */
+	    memset (d, '.', len);
+	    d[len] = 0;
+	    return d;
+	} else {
+	    /* Begin with ... and add the rest of the source string */
+	    memset (d, '.', 3);
+	    strcpy (d + 3, s + 3 + source_len - len);
+	}
     } else
+	/* We can copy the whole line */
 	strcpy (d, s);
     return d;
 }
-#endif
 
+/*
+ * Quote the filename for the purpose of inserting it into the command
+ * line.  If quote_percent is 1, replace "%" with "%%" - the percent is
+ * processed by the mc command line.
+ */
 char *
 name_quote (const char *s, int quote_percent)
 {
     char *ret, *d;
-    
-    d = ret = g_malloc (strlen (s)*2 + 2 + 1);
+
+    d = ret = g_malloc (strlen (s) * 2 + 2 + 1);
     if (*s == '-') {
-        *d++ = '.';
-        *d++ = '/';
+	*d++ = '.';
+	*d++ = '/';
     }
 
     for (; *s; s++, d++) {
-	switch (*s)
-	{
-	    case '%':
-		if (quote_percent)
-		    *d++ = '%';
-		break;
-	    case '\'':
-	    case '\\':
-	    case '\r':
-	    case '\n':
-	    case '\t':
-	    case '"':
-	    case ':':
-	    case ';':
-	    case ' ':
-	    case '?':
-	    case '|':
-	    case '[':
-	    case ']':
-	    case '{':
-	    case '}':
-	    case '<':
-	    case '>':
-	    case '`':
-	    case '~':
-	    case '!':
-	    case '@':
-	    case '#':
-	    case '$':
-	    case '^':
-	    case '&':
-	    case '*':
-	    case '(':
-	    case ')':
+	switch (*s) {
+	case '%':
+	    if (quote_percent)
+		*d++ = '%';
+	    break;
+	case '\'':
+	case '\\':
+	case '\r':
+	case '\n':
+	case '\t':
+	case '"':
+	case ';':
+	case ' ':
+	case '?':
+	case '|':
+	case '[':
+	case ']':
+	case '{':
+	case '}':
+	case '<':
+	case '>':
+	case '`':
+	case '!':
+	case '$':
+	case '&':
+	case '*':
+	case '(':
+	case ')':
+	    *d++ = '\\';
+	    break;
+	case '~':
+	case '#':
+	    if (d == ret)
 		*d++ = '\\';
+	    break;
 	}
 	*d = *s;
     }
@@ -209,45 +202,40 @@ name_quote (const char *s, int quote_percent)
     return ret;
 }
 
-#ifndef VFS_STANDALONE
 char *
 fake_name_quote (const char *s, int quote_percent)
 {
     return g_strdup (s);
 }
 
-/* If passed an empty txt (this usually means that there is an error)
- * in the upper layers, we return "/"
+/*
+ * Remove the middle part of the string to fit given length.
+ * Use "~" to show where the string was truncated.
+ * Return static buffer, no need to free() it.
  */
-char *name_trunc (char *txt, int trunc_len)
+char *
+name_trunc (const char *txt, int trunc_len)
 {
-    static char x [MC_MAXPATHLEN+MC_MAXPATHLEN];
-    int    txt_len;
+    static char x[MC_MAXPATHLEN + MC_MAXPATHLEN];
+    int txt_len;
     char *p;
 
-    if (!txt)
-	txt = PATH_SEP_STR;
-    
-    if (trunc_len > sizeof (x)-1){
-	fprintf (stderr, _("name_trunc: too big"));
-	trunc_len = sizeof (x)-1;
+    if (trunc_len > sizeof (x) - 1) {
+	trunc_len = sizeof (x) - 1;
     }
     txt_len = strlen (txt);
-    if (txt_len <= trunc_len)
+    if (txt_len <= trunc_len) {
 	strcpy (x, txt);
-    else if (tilde_trunc){
-	int y = (trunc_len/2) + (trunc_len % 2);
-	strncpy (x, txt, y);
-	strncpy (x+y, txt+txt_len-(trunc_len/2), trunc_len/2);
-	x [y] = '~';
     } else {
-	strncpy (x, txt, trunc_len-1);
-	x [trunc_len-1] = '>';
+	int y = (trunc_len / 2) + (trunc_len % 2);
+	strncpy (x, txt, y);
+	strncpy (x + y, txt + txt_len - (trunc_len / 2), trunc_len / 2);
+	x[y] = '~';
     }
-    x [trunc_len] = 0;
+    x[trunc_len] = 0;
     for (p = x; *p; p++)
-        if (!is_printable (*p))
-            *p = '?';
+	if (!is_printable (*p))
+	    *p = '?';
     return x;
 }
 
@@ -259,10 +247,10 @@ char *size_trunc (double size)
     
     if (size > 999999999L){
 	divisor = 1024;
-	xtra = "Kb";
+	xtra = "K";
 	if (size/divisor > 999999999L){
 	    divisor = 1024*1024;
-	    xtra = "Mb";
+	    xtra = "M";
 	}
     }
     g_snprintf (x, sizeof (x), "%.0f%s", (size/divisor), xtra);
@@ -279,7 +267,7 @@ char *size_trunc_sep (double size)
     p += strlen (p) - 1;
     d = x + sizeof (x) - 1;
     *d-- = 0;
-    while (p >= y && isalpha (*p))
+    while (p >= y && isalpha ((unsigned char) *p))
 	*d-- = *p--;
     for (count = 0; p >= y; count++){
 	if (count == 3){
@@ -294,6 +282,54 @@ char *size_trunc_sep (double size)
     return d;
 }
 
+/*
+ * Print file SIZE to BUFFER, but don't exceed LEN characters,
+ * not including trailing 0. BUFFER should be at least LEN+1 long.
+ * This function is called for every file on panels, so avoid
+ * floating point by any means.
+ *
+ * Units: size units (filesystem sizes are 1K blocks)
+ *    0=bytes, 1=Kbytes, 2=Mbytes, etc.
+ */
+void
+size_trunc_len (char *buffer, int len, off_t size, int units)
+{
+    /* Avoid taking power for every file.  */
+    static const off_t power10 [] =
+	{1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000,
+	 1000000000};
+    static const char * const suffix [] =
+	{"", "K", "M", "G", "T", "P", "E", "Z", "Y", NULL};
+    int j = 0;
+
+    /* Don't print more than 9 digits - use suffix.  */
+    if (len == 0 || len > 9)
+	len = 9;
+
+    for (j = units; suffix [j] != NULL; j++) {
+	if (size == 0) {
+	    if (j == units) {
+		/* Empty files will print "0" even with minimal width.  */
+		g_snprintf (buffer, len + 1, "0");
+		break;
+	    }
+
+	    /* Use "~K" or just "K" if len is 1.  Use "B" for bytes.  */
+	    g_snprintf (buffer, len + 1, (len > 1) ? "~%s" : "%s",
+			(j > 1) ? suffix[j - 1] : "B");
+	    break;
+	}
+
+	if (size < power10 [len - (j > 0)]) {
+	    g_snprintf (buffer, len + 1, "%lu%s", (unsigned long) size, suffix[j]);
+	    break;
+	}
+
+	/* Powers of 1024, with rounding.  */
+	size = (size + 512) >> 10;
+    }
+}
+
 int is_exe (mode_t mode)
 {
     if ((S_IXUSR & mode) || (S_IXGRP & mode) || (S_IXOTH & mode))
@@ -303,33 +339,56 @@ int is_exe (mode_t mode)
 
 #define ismode(n,m) ((n & m) == m)
 
-char *string_perm (mode_t mode_bits)
+char *
+string_perm (mode_t mode_bits)
 {
-    static char mode [11];
+    static char mode[11];
 
     strcpy (mode, "----------");
-    if (ismode (mode_bits, S_IFDIR)) mode [0] = 'd';
-#ifdef S_IFSOCK
-    if (ismode (mode_bits, S_IFSOCK)) mode [0] = 's';
-#endif
-    if (ismode (mode_bits, S_IXOTH)) mode [9] = 'x';
-    if (ismode (mode_bits, S_IWOTH)) mode [8] = 'w';
-    if (ismode (mode_bits, S_IROTH)) mode [7] = 'r';
-    if (ismode (mode_bits, S_IXGRP)) mode [6] = 'x';
-    if (ismode (mode_bits, S_IWGRP)) mode [5] = 'w';
-    if (ismode (mode_bits, S_IRGRP)) mode [4] = 'r';
-    if (ismode (mode_bits, S_IXUSR)) mode [3] = 'x';
-    if (ismode (mode_bits, S_IWUSR)) mode [2] = 'w';
-    if (ismode (mode_bits, S_IRUSR)) mode [1] = 'r';
-#ifndef OS2_NT
-    if (ismode (mode_bits, S_ISUID)) mode [3] = (mode [3] == 'x') ? 's' : 'S';
-    if (ismode (mode_bits, S_ISGID)) mode [6] = (mode [6] == 'x') ? 's' : 'S';
-    if (ismode (mode_bits, S_IFCHR)) mode [0] = 'c';
-    if (ismode (mode_bits, S_IFBLK)) mode [0] = 'b';
-    if (ismode (mode_bits, S_ISVTX)) mode [9] = (mode [9] == 'x') ? 't' : 'T';
-    if (ismode (mode_bits, S_IFLNK)) mode [0] = 'l';
-    if (ismode (mode_bits, S_IFIFO)) mode [0] = 'p';
-#endif
+    if (S_ISDIR (mode_bits))
+	mode[0] = 'd';
+    if (S_ISCHR (mode_bits))
+	mode[0] = 'c';
+    if (S_ISBLK (mode_bits))
+	mode[0] = 'b';
+    if (S_ISLNK (mode_bits))
+	mode[0] = 'l';
+    if (S_ISFIFO (mode_bits))
+	mode[0] = 'p';
+    if (S_ISSOCK (mode_bits))
+	mode[0] = 's';
+    if (S_ISDOOR (mode_bits))
+	mode[0] = 'D';
+    if (ismode (mode_bits, S_IXOTH))
+	mode[9] = 'x';
+    if (ismode (mode_bits, S_IWOTH))
+	mode[8] = 'w';
+    if (ismode (mode_bits, S_IROTH))
+	mode[7] = 'r';
+    if (ismode (mode_bits, S_IXGRP))
+	mode[6] = 'x';
+    if (ismode (mode_bits, S_IWGRP))
+	mode[5] = 'w';
+    if (ismode (mode_bits, S_IRGRP))
+	mode[4] = 'r';
+    if (ismode (mode_bits, S_IXUSR))
+	mode[3] = 'x';
+    if (ismode (mode_bits, S_IWUSR))
+	mode[2] = 'w';
+    if (ismode (mode_bits, S_IRUSR))
+	mode[1] = 'r';
+#ifdef S_ISUID
+    if (ismode (mode_bits, S_ISUID))
+	mode[3] = (mode[3] == 'x') ? 's' : 'S';
+#endif				/* S_ISUID */
+#ifdef S_ISGID
+    if (ismode (mode_bits, S_ISGID))
+	mode[6] = (mode[6] == 'x') ? 's' : 'S';
+#endif				/* S_ISGID */
+#ifdef S_ISVTX
+    if (ismode (mode_bits, S_ISVTX))
+	mode[9] = (mode[9] == 'x') ? 't' : 'T';
+#endif				/* S_ISVTX */
     return mode;
 }
 
@@ -345,7 +404,7 @@ char *string_perm (mode_t mode_bits)
 char *
 strip_password (char *p, int has_prefix)
 {
-    static struct {
+    static const struct {
 	char *name;
         size_t len;
     } prefixes[] = { {"/#ftp:", 6},
@@ -387,7 +446,7 @@ strip_password (char *p, int has_prefix)
     return (result);
 }
 
-char *strip_home_and_password(char *dir)
+char *strip_home_and_password(const char *dir)
 {
     size_t len;
     static char newdir [MC_MAXPATHLEN];
@@ -395,7 +454,7 @@ char *strip_home_and_password(char *dir)
     if (home_dir && !strncmp (dir, home_dir, len = strlen (home_dir)) && 
 	(dir[len] == PATH_SEP || dir[len] == '\0')){
 	newdir [0] = '~';
-	strcpy (&newdir [1], &dir [strlen (home_dir)]);
+	strcpy (&newdir [1], &dir [len]);
 	return newdir;
     } 
 
@@ -518,33 +577,6 @@ char *extension (char *filename)
     return "";
 }
 
-/* This routine uses the fact that x is at most 14 chars or so */
-char *split_extension (char *x, int pad)
-{
-    return x;
-
-    /* Buggy code 
-    if (!align_extensions)
-	return x;
-
-    if (strlen (x) >= pad)
-	return x;
-    
-    if ((ext = extension (x)) == x || *ext == 0)
-	return x;
-
-    strcpy (xbuf, x);
-    for (i = strlen (x); i < pad; i++)
-	xbuf [i] = ' ';
-    xbuf [pad] = 0;
-
-    l = strlen (ext);
-    for (i = 0; i < l; i++)
-	xbuf [pad-i] = *(ext+l-i-1);
-    for (i = xbuf + (ext - x); i < 
-    return xbuf; */
-}
-
 int get_int (char *file, char *key, int def)
 {
     return GetPrivateProfileInt (app_text, key, def, file);
@@ -570,10 +602,11 @@ char *load_file (char *filename)
     char *data;
     long read_size;
     
-    if (stat (filename, &s) != 0){
+    if ((data_file = fopen (filename, "r")) == NULL){
 	return 0;
     }
-    if ((data_file = fopen (filename, "r")) == NULL){
+    if (fstat (fileno (data_file), &s) != 0){
+	fclose (data_file);
 	return 0;
     }
     data = (char *) g_malloc (s.st_size+1);
@@ -589,6 +622,43 @@ char *load_file (char *filename)
     }
 }
 
+char *load_mc_home_file (const char *filename, char ** allocated_filename)
+{
+    char *hintfile_base, *hintfile;
+    char *lang;
+    char *data;
+
+    hintfile_base = concat_dir_and_file (mc_home, filename);
+    lang = guess_message_value ();
+
+    hintfile = g_strdup_printf ("%s.%s", hintfile_base, lang);
+    data = load_file (hintfile);
+
+    if (!data) {
+	g_free (hintfile);
+	hintfile = g_strdup_printf ("%s.%.2s", hintfile_base, lang);
+	data = load_file (hintfile);
+
+	if (!data) {
+	    g_free (hintfile);
+	    hintfile = hintfile_base;
+	    data = load_file (hintfile_base);
+	}
+    }
+
+    g_free (lang);
+
+    if (hintfile != hintfile_base)
+	g_free (hintfile_base);
+
+    if (allocated_filename)
+	*allocated_filename = hintfile;
+    else
+	g_free (hintfile);
+
+    return data;
+}
+
 /* Check strftime() results. Some systems (i.e. Solaris) have different
 short-month-name sizes for different locales */ 
 size_t i18n_checktimelength (void)
@@ -614,7 +684,8 @@ char *file_date (time_t when)
     static char timebuf [MAX_I18NTIMELENGTH + 1];
     time_t current_time = time ((time_t) 0);
     static size_t i18n_timelength = 0;
-    static char *fmt, *fmtyear, *fmttime;
+    static char *fmtyear, *fmttime;
+    char *fmt;
 
     if (i18n_timelength == 0){
 	i18n_timelength = i18n_checktimelength() + 1;
@@ -642,13 +713,6 @@ char *file_date (time_t when)
     return timebuf;
 }
 
-/* Like file_date, but packs the data to fit in 10 columns */
-char *file_date_pck (time_t when)
-{
-    /* FIXME: Should return only 10 chars, not 14 */
-    return file_date (when);
-}
-
 char *extract_line (char *s, char *top)
 {
     static char tmp_line [BUF_MEDIUM];
@@ -697,44 +761,6 @@ char *x_basename (char *s)
     return ((where = strrchr (s, PATH_SEP))) ? where + 1 : s;
 }
 
-void my_putenv (char *name, char *data)
-{
-    char *full;
-
-    full = malloc (strlen (name) + strlen (data) + 2);
-    strcpy (full, name);
-    strcat (full, "=");
-    strcat (full, data);
-    putenv (full);
-
-    /* WARNING: NEVER FREE THE full VARIABLE!!!!!!!!!!!!!!!!!!!!!!!! */
-    /* It is used by putenv. Freeing it will corrupt the environment */
-}
-
-#if 0
-static void my_putenv_expand (char *name, char macro_code)
-{
-    char *data;
-
-    data = expand_format (macro_code);
-    my_putenv (name, data);
-    g_free (data);
-}
-
-/* Puts some status information in to the environment so that
-   processes to be executed can access it. */
-static void prepare_environment (void)
-{
-    my_putenv_expand ("MC_CURRENT_DIR", 'd');
-    my_putenv_expand ("MC_OTHER_DIR", 'D');
-    my_putenv_expand ("MC_CURRENT_FILE", 'f');
-    my_putenv_expand ("MC_OTHER_FILE", 'F');
-    my_putenv_expand ("MC_CURRENT_TAGGED", 't');
-    my_putenv_expand ("MC_OTHER_TAGGED", 'T');
-    /* MC_CONTROL_FILE has been added to environment on startup */
-}
-#endif
-#endif /* VFS_STANDALONE */
 
 char *unix_error_string (int error_num)
 {
@@ -745,24 +771,6 @@ char *unix_error_string (int error_num)
     return buffer;
 }
 
-#ifndef VFS_STANDALONE	
-long blocks2kilos (int blocks, int bsize)
-{
-    if (bsize > 1024){
-	return blocks * (bsize / 1024);
-    } else if (bsize < 1024){
-	return blocks / (1024 /bsize);
-    } else
-	return blocks;
-}
-
-void init_my_statfs (void)
-{
-#ifndef NO_INFOMOUNT
-    mount_list = read_filesystem_list (1, 1);
-#endif
-}
-
 char *skip_separators (char *s)
 {
     for (;*s; s++)
@@ -774,7 +782,7 @@ char *skip_separators (char *s)
 char *skip_numbers (char *s)
 {
     for (;*s; s++)
-	if (!isdigit (*s))
+	if (!isdigit ((unsigned int) *s))
 	    break;
     return s;
 }
@@ -789,190 +797,130 @@ char *skip_numbers (char *s)
  * terminfo databases, except the Hewlett-Packard 70092 and some Wyse
  * terminals.  If I hear from a single person who uses such a terminal
  * with MC, I'll be glad to add support for it.  (Dugan)
+ * Non-printable characters are also removed.
  */
 
 char *strip_ctrl_codes (char *s)
 {
-    int i;  /* Current length of the string's correct (stripped) prefix */
-    int j;  /* Number of control characters we have skipped so far */
+    char *w; /* Current position where the stripped data is written */
+    char *r; /* Current position where the original data is read */
 
     if (!s)
 	return 0;
-    
-    for (i = 0, j = 0; s [i+j]; ++i)
-	if (s [i+j] != ESC_CHAR){
-	    if (j)
-		s [i] = s [i+j];
-	} else {
-	    ++j;
-	    if (s [i+j++] == '[')
-		while (strchr ("0123456789;?", s [i+j++]))
-		    /* Skip the control sequence's arguments */ ;
-	    --i;
+
+    for (w = s, r = s; *r; ) {
+	if (*r == ESC_CHAR) {
+	    /* Skip the control sequence's arguments */ ;
+	    if (*(++r) == '[') {
+		/* strchr() matches trailing binary 0 */
+		while (*(++r) && strchr ("0123456789;?", *r));
+	    }
+
+	    /*
+	     * Now we are at the last character of the sequence.
+	     * Skip it unless it's binary 0.
+	     */
+	    if (*r)
+		r++;
+	    continue;
 	}
-    s[i] = 0;
+
+	if (is_printable(*r))
+	    *w++ = *r;
+	++r;
+    }
+    *w = 0;
     return s;
 }
 
-#endif /* VFS_STANDALONE */
 
-/*
- * Ok, on systems running glibc, getcwd does not use popen pwd on
- * Linux at least
- */
-#ifdef __GLIBC__
-#  ifdef linux
-#     undef HAVE_GETWD
-#  endif
-#endif
-
-/* getwd is better than getcwd, the later uses a popen ("pwd"); */
+#ifndef USE_VFS
 char *get_current_wd (char *buffer, int size)
 {
     char *p;
+    int len;
 
-#ifdef HAVE_GETWD
-    p = (char *) getwd (buffer);
-#else
-    p = getcwd (buffer, size);
-#endif
-    return p;
+    p = g_get_current_dir ();
+    len = strlen(p) + 1;
+
+    if (len > size) {
+	g_free (p);
+	return NULL;
+    }
+
+    strncpy (buffer, p, len);
+    g_free (p);
+
+    return buffer;
 }
+#endif /* !USE_VFS */
 
-#define CHECK(x) if (x == -1) return 0;
-
-static long
-get_small_endian_long (int fd)
+/* This function returns 0 if the file is not in not compressed by
+ * one of the supported compressors (gzip, bzip, bzip2).  Otherwise,
+ * the compression type is returned, as defined in util.h
+ * Warning: this function moves the current file pointer */
+int get_compression_type (int fd)
 {
-    unsigned char buffer [4];
+    unsigned char magic[4];
 
-    CHECK (mc_read (fd, buffer, 4));
-    return (buffer [3] << 24) | (buffer [2] << 16) | (buffer [1] << 8) | buffer [0];
-}
-
-/*
- * This constant makes the magic array on the stack be larger than
- * it needs because Linux when reading the second byte of /proc/locks
- * for example will write 2 bytes, even if we only asked for one
- */
-#define LINUX_HAS_PROBLEMS_WHEN_READING_PROC_LOCKS_ON_SOME_KERNELS 40
-
-/* This function returns 0 if the file is not in gunzip format  */
-/* or how much memory must be allocated to load the gziped file */
-/* Warning: this function moves the current file pointer */
-long int is_gunzipable (int fd, int *type)
-{
-    unsigned char magic [4+LINUX_HAS_PROBLEMS_WHEN_READING_PROC_LOCKS_ON_SOME_KERNELS];
-    
-    *type = ISGUNZIPABLE_GUNZIP;
-	
     /* Read the magic signature */
-    CHECK (mc_read (fd, &magic [0], 4));
-	
+    if (mc_read (fd, (char *) magic, 4) != 4)
+	return COMPRESSION_NONE;
+
     /* GZIP_MAGIC and OLD_GZIP_MAGIC */
-    if (magic [0] == 037 && (magic [1] == 0213 || magic [1] == 0236)){
-	/* Read the uncompressed size of the file */
-	mc_lseek (fd, -4, SEEK_END);
-	return get_small_endian_long (fd);
+    if (magic[0] == 037 && (magic[1] == 0213 || magic[1] == 0236)) {
+	return COMPRESSION_GZIP;
     }
 
     /* PKZIP_MAGIC */
-    if (magic [0] == 0120 && magic [1] == 0113 && magic [2] == 003 && magic [3] == 004){
+    if (magic[0] == 0120 && magic[1] == 0113 && magic[2] == 003
+	&& magic[3] == 004) {
 	/* Read compression type */
 	mc_lseek (fd, 8, SEEK_SET);
-	CHECK (mc_read (fd, &magic [0], 2));
-	
+	if (mc_read (fd, (char *) magic, 2) != 2)
+	    return COMPRESSION_NONE;
+
 	/* Gzip can handle only deflated (8) or stored (0) files */
-	if ((magic [0] != 8 && magic [0] != 0) || magic [1] != 0)
-	     return 0;
-        /* Read the uncompressed size of the first file in the archive */
-	mc_lseek (fd, 22, SEEK_SET);
-	return get_small_endian_long (fd);
+	if ((magic[0] != 8 && magic[0] != 0) || magic[1] != 0)
+	    return COMPRESSION_NONE;
+
+	/* Compatible with gzip */
+	return COMPRESSION_GZIP;
     }
 
     /* PACK_MAGIC and LZH_MAGIC and compress magic */
-    if (magic [0] == 037 && (magic [1] ==  036 || magic [1] == 0240 || magic [1] == 0235)){
-	 /* In case the file is packed, sco lzhed or compress_magic, the */
-	 /* program guesses that the uncompressed size is (at most) four */
-	 /* times the length of the compressed size, if the compression  */
-	 /* ratio is more than 4:1 the end of the file is not displayed  */
-	 return 4*mc_lseek (fd, 0, SEEK_END);
+    if (magic[0] == 037
+	&& (magic[1] == 036 || magic[1] == 0240 || magic[1] == 0235)) {
+	/* Compatible with gzip */
+	return COMPRESSION_GZIP;
     }
 
     /* BZIP and BZIP2 files */
     if ((magic[0] == 'B') && (magic[1] == 'Z') &&
-	(magic [3] >= '1') && (magic [3] <= '9')){
-            switch (magic[2]) {
-                case '0':
-                    *type = ISGUNZIPABLE_BZIP;
-                    return 5*mc_lseek (fd, 0, SEEK_END);
-                case 'h': 
-	            *type = ISGUNZIPABLE_BZIP2;
-	            return 5*mc_lseek (fd, 0, SEEK_END);
-            }
+	(magic[3] >= '1') && (magic[3] <= '9')) {
+	switch (magic[2]) {
+	case '0':
+	    return COMPRESSION_BZIP;
+	case 'h':
+	    return COMPRESSION_BZIP2;
+	}
     }
     return 0;
 }
 
-char *
+const char *
 decompress_extension (int type)
 {
 	switch (type){
-	case ISGUNZIPABLE_GUNZIP: return "#ugz";
-	case ISGUNZIPABLE_BZIP:   return "#ubz";
-	case ISGUNZIPABLE_BZIP2:  return "#ubz2";
+	case COMPRESSION_GZIP: return "#ugz";
+	case COMPRESSION_BZIP:   return "#ubz";
+	case COMPRESSION_BZIP2:  return "#ubz2";
 	}
 	/* Should never reach this place */
 	fprintf (stderr, "Fatal: decompress_extension called with an unknown argument\n");
 	return 0;
 }
 
-char *
-decompress_command (int type)
-{
-	switch (type){
-	case ISGUNZIPABLE_GUNZIP:
-		return "gzip -cdf";
-		
-	case ISGUNZIPABLE_BZIP:
-		return "bzip -d";
-		
-	case ISGUNZIPABLE_BZIP2:
-		return "bzip2 -d";
-	}
-	/* Should never reach this place */
-	fprintf (stderr, "Fatal: decompress_command called with an unknown argument\n");
-	return 0;
-}
-
-void
-decompress_command_and_arg (int type, char **cmd, char **flags)
-{
-	switch (type){
-	case ISGUNZIPABLE_GUNZIP:
-		*cmd   = "gzip";
-		*flags = "-cdf";
-		return;
-
-	case ISGUNZIPABLE_BZIP:
-		*cmd   = "bzip";
-		*flags = "-d";
-		return;
-
-		
-	case ISGUNZIPABLE_BZIP2:
-		*cmd   = "bzip2";
-		*flags = "-d";
-		return;
-	}
-	*cmd   = 0;
-	*flags = 0;
-	
-	/* Should never reach this place */
-	fprintf (stderr, "Fatal: decompress_command called with an unknown argument\n");
-}
-
-#ifndef VFS_STANDALONE
 /* Hooks */
 void add_hook (Hook **hook_list, void (*hook_fn)(void *), void *data)
 {
@@ -1080,27 +1028,11 @@ char *convert_controls (char *s)
 		*q++ = *p++;
 	}
     }
-    *q++ = 0;
+    *q = 0;
     return valcopy;
 }
 
-/* Reverse the string */
-char *reverse_string (char *string)
-{
-    int len = strlen (string);
-    int i;
-    const int steps = len/2;
-    
-    for (i = 0; i < steps; i++){
-	char c = string [i];
-    
-	string [i] = string [len-i-1];
-	string [len-i-1] = c;
-    }
-    return string;
-}
-
-char *resolve_symlinks (char *path)
+static char *resolve_symlinks (char *path)
 {
     char *buf, *buf2, *p, *q, *r, c;
     int len;
@@ -1221,45 +1153,6 @@ char *diff_two_paths (char *first, char *second)
     return buf;
 }
 
-#ifndef HAVE_TRUNCATE
-/* On SCO and Windows NT systems */
-int my_ftruncate (int fd, long size)
-{
-#ifdef OS2_NT
-    if(_chsize(fd, size))
-	return -1;
-    else 
-	return 0;
-#else
-    struct flock lk;
-    
-    lk.l_whence = 0;
-    lk.l_start = size;
-    lk.l_len = 0;
-    
-    return fcntl (fd, F_FREESP, &lk);
-#endif
-}
-
-int truncate (const char *path, long size)
-{
-    int fd;
-    int res;
-    
-    fd = open (path, O_RDWR, 0);
-    if (fd < 0)
-	return fd;
-    res = my_ftruncate (fd, size);
-    if (res < 0)
-	return res;
-    close (fd);
-    return 0;
-
-}
-
-#endif
-#endif /* VFS_STANDALONE */
-
 /* If filename is NULL, then we just append PATH_SEP to the dir */
 char *
 concat_dir_and_file (const char *dir, const char *file)
@@ -1272,3 +1165,182 @@ concat_dir_and_file (const char *dir, const char *file)
 	return  g_strconcat (dir, PATH_SEP_STR, file, NULL);
 }
 
+/* Following code heavily borrows from libiberty, mkstemps.c */
+
+/* Number of attempts to create a temporary file */
+#ifndef TMP_MAX
+#define TMP_MAX 16384
+#endif /* !TMP_MAX */
+
+/*
+ * Arguments:
+ * pname (output) - pointer to the name of the temp file (needs g_free).
+ *                  NULL if the function fails.
+ * prefix - part of the filename before the random part.
+ *          Prepend $TMPDIR or /tmp if there are no path separators.
+ * suffix - if not NULL, part of the filename after the random part.
+ *
+ * Result:
+ * handle of the open file or -1 if couldn't open any.
+ */
+int
+mc_mkstemps (char **pname, const char *prefix, const char *suffix)
+{
+    static const char letters[]
+	= "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
+    static unsigned long value;
+    struct timeval tv;
+    char *tmpbase;
+    char *tmpname;
+    char *XXXXXX;
+    int count;
+
+    if (strchr (prefix, PATH_SEP) == NULL) {
+	/* Add prefix first to find the position of XXXXXX */
+	tmpbase = concat_dir_and_file (mc_tmpdir (), prefix);
+    } else {
+	tmpbase = g_strdup (prefix);
+    }
+
+    tmpname = g_strconcat (tmpbase, "XXXXXX", suffix, NULL);
+    *pname = tmpname;
+    XXXXXX = &tmpname[strlen (tmpbase)];
+    g_free (tmpbase);
+
+    /* Get some more or less random data.  */
+    gettimeofday (&tv, NULL);
+    value += (tv.tv_usec << 16) ^ tv.tv_sec ^ getpid ();
+
+    for (count = 0; count < TMP_MAX; ++count) {
+	unsigned long v = value;
+	int fd;
+
+	/* Fill in the random bits.  */
+	XXXXXX[0] = letters[v % 62];
+	v /= 62;
+	XXXXXX[1] = letters[v % 62];
+	v /= 62;
+	XXXXXX[2] = letters[v % 62];
+	v /= 62;
+	XXXXXX[3] = letters[v % 62];
+	v /= 62;
+	XXXXXX[4] = letters[v % 62];
+	v /= 62;
+	XXXXXX[5] = letters[v % 62];
+
+	fd = open (tmpname, O_RDWR | O_CREAT | O_TRUNC | O_EXCL,
+		   S_IRUSR | S_IWUSR);
+	if (fd >= 0) {
+	    /* Successfully created.  */
+	    return fd;
+	}
+
+	/* This is a random value.  It is only necessary that the next
+	   TMP_MAX values generated by adding 7777 to VALUE are different
+	   with (module 2^32).  */
+	value += 7777;
+    }
+
+    /* Unsuccessful. Free the filename. */
+    g_free (tmpname);
+    *pname = NULL;
+
+    return -1;
+}
+
+
+/*
+ * Read and restore position for the given filename.
+ * If there is no stored data, return line 1 and col 0.
+ */
+void
+load_file_position (char *filename, long *line, long *column)
+{
+    char *fn;
+    FILE *f;
+    char buf[MC_MAXPATHLEN + 20];
+    int len;
+
+    /* defaults */
+    *line = 1;
+    *column = 0;
+
+    /* open file with positions */
+    fn = concat_dir_and_file (home_dir, MC_FILEPOS);
+    f = fopen (fn, "r");
+    g_free (fn);
+    if (!f)
+	return;
+
+    len = strlen (filename);
+
+    while (fgets (buf, sizeof (buf), f)) {
+	char *p;
+
+	/* check if the filename matches the beginning of string */
+	if (strncmp (buf, filename, len) != 0)
+	    continue;
+
+	/* followed by single space */
+	if (buf[len] != ' ')
+	    continue;
+
+	/* and string without spaces */
+	p = &buf[len + 1];
+	if (strchr (p, ' '))
+	    continue;
+
+	*line = atol (p);
+	p = strchr (buf, ';');
+	*column = atol (&p[1]);
+    }
+    fclose (f);
+}
+
+/* Save position for the given file */
+void
+save_file_position (char *filename, long line, long column)
+{
+    char *tmp, *fn;
+    FILE *f, *t;
+    char buf[MC_MAXPATHLEN + 20];
+    int i = 1;
+    int len;
+
+    len = strlen (filename);
+
+    tmp = concat_dir_and_file (home_dir, MC_FILEPOS_TMP);
+    fn = concat_dir_and_file (home_dir, MC_FILEPOS);
+
+    /* open temporary file */
+    t = fopen (tmp, "w");
+    if (!t) {
+	g_free (tmp);
+	g_free (fn);
+	return;
+    }
+
+    /* put the new record */
+    fprintf (t, "%s %ld;%ld\n", filename, line, column);
+
+    /* copy records from the old file */
+    f = fopen (fn, "r");
+    if (f) {
+	while (fgets (buf, sizeof (buf), f)) {
+	    /* Skip entries for the current filename */
+	    if (strncmp (buf, filename, len) == 0 && buf[len] == ' '
+		&& !strchr (&buf[len + 1], ' '))
+		continue;
+
+	    fprintf (t, "%s", buf);
+	    if (++i > MC_FILEPOS_ENTRIES)
+		break;
+	}
+	fclose (f);
+    }
+
+    fclose (t);
+    rename (tmp, fn);
+    g_free (tmp);
+    g_free (fn);
+}
