@@ -6,8 +6,6 @@
    Written by: 1998 Pavel Machek
    Spaces fix: 2000 Michal Svec
 
-   $Id: fish.c,v 1.30 2001/03/08 01:04:58 proskin Exp $
-
    Derived from ftpfs.c.
    
    This program is free software; you can redistribute it and/or
@@ -36,6 +34,8 @@
 
 /* Define this if your ssh can take -I option */
 
+#include <config.h>
+
 #undef HAVE_HACKED_SSH
 
 #include "utilvfs.h"
@@ -44,7 +44,6 @@
 #include "xdirentry.h"
 #include "vfs.h"
 #include "tcputil.h"
-#include "container.h"
 #include "fish.h"
 
 /*
@@ -67,6 +66,10 @@ static int force_expiration = 0;
 #define WANT_STRING 0x02
 static char reply_str [80];
 
+static int
+command (vfs *me, vfs_s_super *super, int wait_reply, const char *fmt, ...)
+    __attribute__ ((format (printf, 4, 5)));
+
 static int decode_reply (char *s, int was_garbage)
 {
     int code;
@@ -102,7 +105,8 @@ static int get_reply (vfs *me, int sock, char *string_buf, int string_len)
 
 #define SUP super->u.fish
 
-static int command (vfs *me, vfs_s_super *super, int wait_reply, char *fmt, ...)
+static int
+command (vfs *me, vfs_s_super *super, int wait_reply, const char *fmt, ...)
 {
     va_list ap;
     char *str;
@@ -138,16 +142,15 @@ free_archive (vfs *me, vfs_s_super *super)
 {
     if ((SUP.sockw != -1) || (SUP.sockr != -1)){
 	print_vfs_message (_("fish: Disconnecting from %s"), super->name?super->name:"???");
-	command(me, super, NONE, "#BYE\nlogout\n");
+	command(me, super, NONE, "#BYE\nexit\n");
 	close(SUP.sockw);
 	close(SUP.sockr);
 	SUP.sockw = SUP.sockr = -1;
     }
-    ifree (SUP.host);
-    ifree (SUP.home);
-    ifree (SUP.user);
-    ifree (SUP.cwdir);
-    ifree (SUP.password);
+    g_free (SUP.host);
+    g_free (SUP.user);
+    g_free (SUP.cwdir);
+    g_free (SUP.password);
 }
 
 static void
@@ -157,10 +160,10 @@ pipeopen(vfs_s_super *super, char *path, char *argv[])
     int res;
 
     if ((pipe(fileset1)<0) || (pipe(fileset2)<0)) 
-	vfs_die("Could not pipe(): %m.");
+	vfs_die("Cannot pipe(): %m.");
     
     if ((res = fork())) {
-        if (res<0) vfs_die("Could not fork(): %m.");
+        if (res<0) vfs_die("Cannot fork(): %m.");
 	/* We are the parent */
 	close(fileset1[0]);
 	SUP.sockw = fileset1[1];
@@ -172,10 +175,11 @@ pipeopen(vfs_s_super *super, char *path, char *argv[])
 	close(fileset1[0]); close(fileset1[1]);
 	close(1); close(2);
 	dup(fileset2[1]);
-	dup(fileset2[1]);
+	/* stderr to /dev/null */
+	open ("/dev/null", O_WRONLY);
 	close(fileset2[0]); close(fileset2[1]);
 	execvp(path, argv);
-	vfs_die("Exec failed.");
+	_exit(3);
     }
 }
 
@@ -225,7 +229,7 @@ open_archive_int (vfs *me, vfs_s_super *super)
        /dev/tty, not from stdin :-(. */
 
 #ifndef HAVE_HACKED_SSH
-	    message_1s (1, MSG_ERROR, _("Sorry, we can not do password authenticated connections for now."));
+	    message_1s (1, MSG_ERROR, _("Sorry, we cannot do password authenticated connections for now."));
 	    ERRNOR (EPERM, -1);
 #endif
 	    if (!SUP.password){
@@ -246,16 +250,28 @@ open_archive_int (vfs *me, vfs_s_super *super)
     }
 
     print_vfs_message( _("fish: Sending initial line...") );
-    if (command (me, super, WAIT_REPLY, "#FISH\necho; start_fish_server; echo '### 200'\n") != COMPLETE)
+    /*
+     * Run `start_fish_server'. If it doesn't exist - no problem,
+     * we'll talk directly to the shell.
+     */
+    if (command (me, super, WAIT_REPLY,
+                 "#FISH\necho; start_fish_server 2>&1;"
+                 " echo '### 200'\n") != COMPLETE)
         ERRNOR (E_PROTO, -1);
 
     print_vfs_message( _("fish: Handshaking version...") );
     if (command (me, super, WAIT_REPLY, "#VER 0.0.0\necho '### 000'\n") != COMPLETE)
         ERRNOR (E_PROTO, -1);
 
+    /* Set up remote locale to C, otherwise dates cannot be recognized */
+    if (command (me, super, WAIT_REPLY, "LANG=C; LC_ALL=C; LC_TIME=C\n"
+					"export LANG; export LC_ALL; export LC_TIME\n"
+					"echo '### 200'\n") != COMPLETE)
+        ERRNOR (E_PROTO, -1);
+
     print_vfs_message( _("fish: Setting up current directory...") );
-    SUP.home = fish_getcwd (me, super);
-    print_vfs_message( _("fish: Connected, home %s."), SUP.home );
+    SUP.cwdir = fish_getcwd (me, super);
+    print_vfs_message( _("fish: Connected, home %s."), SUP.cwdir );
 #if 0
     super->name = g_strconcat ( "/#sh:", SUP.user, "@", SUP.host, "/", NULL );
 #endif
@@ -281,7 +297,7 @@ open_archive (vfs *me, vfs_s_super *super, char *archive_name, char *op)
     SUP.flags = flags;
     if (!strncmp( op, "rsh:", 4 ))
 	SUP.flags |= FISH_FLAG_RSH;
-    SUP.home = NULL;
+    SUP.cwdir = NULL;
     if (password)
 	SUP.password = password;
     return open_archive_int (me, super);
@@ -307,18 +323,6 @@ archive_same(vfs *me, vfs_s_super *super, char *archive_name, char *op, void *co
     return flags;
 }
 
-int
-fish_which (vfs *me, char *path)
-{
-    if (!strncmp (path, "/#sh:", 5))
-        return 1;
-    if (!strncmp (path, "/#ssh:", 6))
-        return 1;
-    if (!strncmp (path, "/#rsh:", 6))
-        return 1;
-    return 0;
-}
-
 static int
 dir_uptodate(vfs *me, vfs_s_inode *ino)
 {
@@ -341,6 +345,7 @@ dir_load(vfs *me, vfs_s_inode *dir, char *remote_path)
     char buffer[8192];
     vfs_s_entry *ent = NULL;
     FILE *logfile;
+    char *quoted_path;
 
     logfile = MEDATA->logfile;
 
@@ -349,22 +354,22 @@ dir_load(vfs *me, vfs_s_inode *dir, char *remote_path)
     gettimeofday(&dir->u.fish.timestamp, NULL);
     dir->u.fish.timestamp.tv_sec += 10; /* was 360: 10 is good for
 					   stressing direntry layer a bit */
-
+    quoted_path = name_quote (remote_path, 0);
     command(me, super, NONE,
 	    "#LIST /%s\n"
-	    "ls -lLa \"/%s\" | grep '^[^cbt]' | (\n"
+	    "ls -lLa /%s 2>/dev/null | grep '^[^cbt]' | (\n"
 	      "while read p x u g s m d y n; do\n"
 	        "echo \"P$p $u.$g\nS$s\nd$m $d $y\n:$n\n\"\n"
 	      "done\n"
 	    ")\n"
-	    "ls -lLa \"/%s\" | grep '^[cb]' | (\n"
+	    "ls -lLa /%s 2>/dev/null | grep '^[cb]' | (\n"
 	      "while read p x u g a i m d y n; do\n"
 	        "echo \"P$p $u.$g\nE$a$i\nd$m $d $y\n:$n\n\"\n"
 	      "done\n"
 	    ")\n"
 	    "echo '### 200'\n",
-	    remote_path, remote_path, remote_path);
-
+	    remote_path, quoted_path, quoted_path);
+    g_free (quoted_path);
 #define SIMPLE_ENTRY vfs_s_generate_entry(me, NULL, dir, 0)
     ent = SIMPLE_ENTRY;
     while (1) {
@@ -398,7 +403,7 @@ dir_load(vfs *me, vfs_s_inode *dir, char *remote_path)
 			  break;  /* We'll do . and .. ourself */
 		      ent->name = g_strdup(buffer+1); 
 		      /* if ((c=strchr(ent->name, ' ')))
-			  *c = 0; / * this is ugly, but we can not handle " " in name */
+			  *c = 0; / * this is ugly, but we cannot handle " " in name */
 		      break;
 	          }
 	case 'S': ST.st_size = atoi(buffer+1); break;
@@ -444,49 +449,80 @@ dir_load(vfs *me, vfs_s_inode *dir, char *remote_path)
     
     vfs_s_free_entry (me, ent);
     me->verrno = E_REMOTE;
-    if (decode_reply(buffer+4, 0) == COMPLETE)
-        return 0;
+    if (decode_reply(buffer+4, 0) == COMPLETE) {
+	g_free (SUP.cwdir);
+	SUP.cwdir = g_strdup (remote_path);
+	print_vfs_message (_("%s: done."), me->name);
+	return 0;
+    }
 
 error:
-    print_vfs_message(_("fish: failed"));
+    print_vfs_message (_("%s: failure"), me->name);
     return 1;
 }
 
 static int
-file_store(vfs *me, vfs_s_super *super, char *name, char *localname)
+file_store(vfs *me, vfs_s_fh *fh, char *name, char *localname)
 {
+    vfs_s_super *super = FH_SUPER;
     int n, total;
     char buffer[8192];
     struct stat s;
     int was_error = 0;
     int h;
+    char *quoted_name;
 
-    h = open(localname, O_RDONLY);
+    h = open (localname, O_RDONLY);
 
-    if (fstat(h, &s)<0)
+    if (h == -1)
 	ERRNOR (EIO, -1);
-
+    if (fstat(h, &s)<0) {
+	close (h);
+	ERRNOR (EIO, -1);
+    }
     /* Use this as stor: ( dd block ; dd smallblock ) | ( cat > file; cat > /dev/null ) */
 
     print_vfs_message(_("fish: store %s: sending command..."), name );
-    if (command (me, super, WAIT_REPLY,
-		 "#STOR %d /%s\n"
-		 "> \"/%s\"\n"
+    quoted_name = name_quote (name, 0);
+    /*
+     * FIXME: Limit size to unsigned long for now.
+     * Files longer than 256 * ULONG_MAX are not supported.
+     */
+    if (!fh->u.fish.append)
+	n = command (me, super, WAIT_REPLY,
+		 "#STOR %lu /%s\n"
+		 "> /%s\n"
 		 "echo '### 001'\n"
 		 "(\n"
-		   "dd bs=4096 count=%d\n"
-		   "dd bs=%d count=1\n"
+		   "dd ibs=256 obs=4096 count=%lu\n"
+		   "dd bs=%lu count=1\n"
 		 ") 2>/dev/null | (\n"
-		   "cat > \"/%s\"\n"
+		   "cat > /%s\n"
 		   "cat > /dev/null\n"
 		 "); echo '### 200'\n",
-	      /* ")\n"                       Why can't it be like this?
-	         "echo '### 200'\n", */
-		 s.st_size, name, name,
-		 s.st_size / 4096, s.st_size % 4096, name)
-	!= PRELIM) 
-        ERRNOR(E_REMOTE, -1);
+		 (unsigned long) s.st_size, name, quoted_name,
+		 (unsigned long) (s.st_size >> 8),
+		 ((unsigned long) s.st_size) & (256 - 1), quoted_name);
+    else
+	n = command (me, super, WAIT_REPLY,
+		 "#STOR %lu /%s\n"
+		 "echo '### 001'\n"
+		 "(\n"
+		   "dd ibs=256 obs=4096 count=%lu\n"
+		   "dd bs=%lu count=1\n"
+		 ") 2>/dev/null | (\n"
+		   "cat >> /%s\n"
+		   "cat > /dev/null\n"
+		 "); echo '### 200'\n",
+		 (unsigned long) s.st_size, name,
+		 (unsigned long) (s.st_size >> 8),
+		 ((unsigned long) s.st_size) & (256 - 1), quoted_name);
 
+    g_free (quoted_name);
+    if (n != PRELIM) {
+	close (h);
+        ERRNOR(E_REMOTE, -1);
+    }
     total = 0;
     
     while (1) {
@@ -505,12 +541,13 @@ file_store(vfs *me, vfs_s_super *super, char *name, char *localname)
 	}
 	disable_interrupt_key();
 	total += n;
-	print_vfs_message(_("fish: storing %s %d (%d)"), 
-			  was_error ? _("zeros") : _("file"), total, s.st_size);
+	print_vfs_message(_("fish: storing %s %d (%lu)"), 
+			  was_error ? _("zeros") : _("file"), total,
+			  (unsigned long) s.st_size);
     }
+    close(h);
     if ((get_reply (me, SUP.sockr, NULL, 0) != COMPLETE) || was_error)
         ERRNOR (E_REMOTE, -1);
-    close(h);
     return 0;
 error_return:
     close(h);
@@ -521,20 +558,25 @@ error_return:
 static int linear_start(vfs *me, vfs_s_fh *fh, int offset)
 {
     char *name;
+    char *quoted_name;
     if (offset)
         ERRNOR (E_NOTSUPP, 0);
 /*    fe->local_stat.st_mtime = 0; FIXME: what is this good for? */
     name = vfs_s_fullpath (me, fh->ino);
     if (!name)
 	return 0;
+    quoted_name = name_quote (name, 0);
+    g_free (name);
+    name = quoted_name;
+    fh->u.fish.append = 0;
     offset = command(me, FH_SUPER, WANT_STRING,
 		"#RETR /%s\n"
-		"ls -l \"/%s\" | (\n"
+		"ls -l /%s 2>/dev/null | (\n"
 		  "read var1 var2 var3 var4 var5 var6\n"
 		  "echo \"$var5\"\n"
 		")\n"
 		"echo '### 100'\n"
-		"cat \"/%s\"\n"
+		"cat /%s\n"
 		"echo '### 200'\n", 
 		name, name, name );
     g_free (name);
@@ -592,7 +634,9 @@ static void
 linear_close (vfs *me, vfs_s_fh *fh)
 {
     if (fh->u.fish.total != fh->u.fish.got)
-        linear_abort(me, fh);
+	linear_abort(me, fh);
+    else if (stat (fh->ino->localname, &fh->ino->u.fish.local_stat) < 0)
+	fh->ino->u.fish.local_stat.st_mtime = 0;
 }
 
 static int
@@ -637,16 +681,20 @@ send_fish_command(vfs *me, vfs_s_super *super, char *cmd, int flags)
     char *rpath; \
     vfs_s_super *super; \
     if (!(rpath = vfs_s_get_path_mangle(me, path, &super, 0))) \
-	return -1;
+	return -1; \
+    rpath = name_quote (rpath, 0);
 
 #define POSTFIX(flags) \
+    g_free (rpath); \
     return send_fish_command(me, super, buf, flags);
 
 static int
 fish_chmod (vfs *me, char *path, int mode)
 {
     PREFIX
-    g_snprintf(buf, sizeof(buf), "#CHMOD %4.4o /%s\nchmod %4.4o \"/%s\"; echo '### 000'\n", 
+    g_snprintf(buf, sizeof(buf), "#CHMOD %4.4o /%s\n"
+				 "chmod %4.4o \"/%s\" 2>/dev/null\n"
+				 "echo '### 000'\n", 
 	    mode & 07777, rpath,
 	    mode & 07777, rpath);
     POSTFIX(OPT_FLUSH);
@@ -656,28 +704,38 @@ fish_chmod (vfs *me, char *path, int mode)
 static int fish_##name (vfs *me, char *path1, char *path2) \
 { \
     char buf[BUF_LARGE]; \
-    char *rpath1 = NULL, *rpath2 = NULL; \
+    char *rpath1, *rpath2; \
     vfs_s_super *super1, *super2; \
     if (!(rpath1 = vfs_s_get_path_mangle(me, path1, &super1, 0))) \
 	return -1; \
     if (!(rpath2 = vfs_s_get_path_mangle(me, path2, &super2, 0))) \
 	return -1; \
-    g_snprintf(buf, 1023, string "\n", rpath1, rpath2, rpath1, rpath2 ); \
+    rpath1 = name_quote (rpath1, 0); \
+    rpath2 = name_quote (rpath2, 0); \
+    g_snprintf(buf, sizeof(buf), string "\n", rpath1, rpath2, rpath1, rpath2); \
+    g_free (rpath1); \
+    g_free (rpath2); \
     return send_fish_command(me, super2, buf, OPT_FLUSH); \
 }
 
 #define XTEST if (bucket1 != bucket2) { ERRNOR (EXDEV, -1); }
-FISH_OP(rename, XTEST, "#RENAME /%s /%s\nmv \"/%s\" \"/%s\"; echo '### 000'" );
-FISH_OP(link,   XTEST, "#LINK /%s /%s\nln \"/%s\" \"/%s\"; echo '### 000'" );
+FISH_OP(rename, XTEST, "#RENAME /%s /%s\n"
+		       "mv /%s /%s 2>/dev/null\n"
+		       "echo '### 000'" )
+FISH_OP(link,   XTEST, "#LINK /%s /%s\n"
+		       "ln /%s /%s 2>/dev/null\n"
+		       "echo '### 000'" )
 
 static int fish_symlink (vfs *me, char *setto, char *path)
 {
     PREFIX
+    setto = name_quote (setto, 0);
     g_snprintf(buf, sizeof(buf),
             "#SYMLINK %s /%s\n"
-	    "ln -s \"%s\" \"/%s\"\n"
+	    "ln -s %s /%s 2>/dev/null\n"
 	    "echo '### 000'\n",
 	    setto, rpath, setto, rpath);
+    g_free (setto);
     POSTFIX(OPT_FLUSH);
 }
 
@@ -685,20 +743,29 @@ static int
 fish_chown (vfs *me, char *path, int owner, int group)
 {
     char *sowner, *sgroup;
+    struct passwd *pw;
+    struct group *gr;
     PREFIX
-    sowner = getpwuid( owner )->pw_name;
-    sgroup = getgrgid( group )->gr_name;
+
+    if ((pw = getpwuid (owner)) == NULL)
+	return 0;
+
+    if ((gr = getgrgid (group)) == NULL)
+	return 0;
+
+    sowner = pw->pw_name;
+    sgroup = gr->gr_name;
     g_snprintf(buf, sizeof(buf),
             "#CHOWN /%s /%s\n"
-	    "chown %s \"/%s\"\n"
+	    "chown %s /%s 2>/dev/null\n"
 	    "echo '### 000'\n", 
 	    sowner, rpath,
 	    sowner, rpath);
     send_fish_command(me, super, buf, OPT_FLUSH); 
-                  /* FIXME: what should we report if chgrp succeeds but chown fails? */
+    /* FIXME: what should we report if chgrp succeeds but chown fails? */
     g_snprintf(buf, sizeof(buf),
             "#CHGRP /%s /%s\n"
-	    "chgrp %s \"/%s\"\n"
+	    "chgrp %s /%s 2>/dev/null\n"
 	    "echo '### 000'\n", 
 	    sgroup, rpath,
 	    sgroup, rpath);
@@ -711,7 +778,7 @@ static int fish_unlink (vfs *me, char *path)
     PREFIX
     g_snprintf(buf, sizeof(buf),
             "#DELE /%s\n"
-	    "rm -f \"/%s\"\n"
+	    "rm -f /%s 2>/dev/null\n"
 	    "echo '### 000'\n",
 	    rpath, rpath);
     POSTFIX(OPT_FLUSH);
@@ -722,7 +789,7 @@ static int fish_mkdir (vfs *me, char *path, mode_t mode)
     PREFIX
     g_snprintf(buf, sizeof(buf),
             "#MKD /%s\n"
-	    "mkdir \"/%s\"\n"
+	    "mkdir /%s 2>/dev/null\n"
 	    "echo '### 000'\n",
 	    rpath, rpath);
     POSTFIX(OPT_FLUSH);
@@ -733,7 +800,7 @@ static int fish_rmdir (vfs *me, char *path)
     PREFIX
     g_snprintf(buf, sizeof(buf),
             "#RMD /%s\n"
-	    "rmdir \"/%s\"\n"
+	    "rmdir /%s 2>/dev/null\n"
 	    "echo '### 000'\n",
 	    rpath, rpath);
     POSTFIX(OPT_FLUSH);
@@ -741,6 +808,18 @@ static int fish_rmdir (vfs *me, char *path)
 
 static int fish_fh_open (vfs *me, vfs_s_fh *fh, int flags, int mode)
 {
+    fh->u.fish.append = 0;
+    /* File will be written only, so no need to retrieve it */
+    if (((flags & O_WRONLY) == O_WRONLY) && !(flags & (O_RDONLY|O_RDWR))){
+	fh->u.fish.append = flags & O_APPEND;
+	if (!fh->ino->localname){
+	    int tmp_handle = mc_mkstemps (&fh->ino->localname, me->name, NULL);
+	    if (tmp_handle == -1)
+		return -1;
+	    close (tmp_handle);
+	}
+	return 0;
+    }
     if (!fh->ino->localname)
 	if (vfs_s_retrieve_file (me, fh->ino)==-1)
 	    return -1;
@@ -777,16 +856,47 @@ static struct vfs_s_data fish_data = {
     linear_close
 };
 
+static void
+fish_fill_names (vfs *me, void (*func)(char *))
+{
+    struct vfs_s_super * super = fish_data.supers;
+    char *flags;
+    char *name;
+    
+    while (super){
+	switch (SUP.flags & (FISH_FLAG_RSH | FISH_FLAG_COMPRESSED)) {
+	case FISH_FLAG_RSH:
+		flags = ":r";
+		break;
+	case FISH_FLAG_COMPRESSED:
+		flags = ":C";
+		break;
+	case FISH_FLAG_RSH | FISH_FLAG_COMPRESSED:
+		flags = "";
+		break;
+	default:
+		flags = "";
+		break;
+	}
+
+	name = g_strconcat ("/#sh:", SUP.user, "@", SUP.host, flags,
+			    "/", SUP.cwdir, NULL);
+	(*func)(name);
+	g_free (name);
+	super = super->next;
+    }
+}
+
 vfs vfs_fish_ops = {
     NULL,	/* This is place of next pointer */
-    "FIles tranferred over SHell",
+    "fish",
     F_EXEC,	/* flags */
     "sh:",	/* prefix */
     &fish_data,	/* data */
     0,		/* errno */
     NULL,
     NULL,
-    vfs_s_fill_names,
+    fish_fill_names,
     NULL,
 
     vfs_s_open,
