@@ -5,8 +5,6 @@
                1995 Jakub Jelinek
 	       1998 Pavel Machek
    
-   $Id: vfs.c,v 1.68 2001/03/05 06:00:27 andrew Exp $
-
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public License
    as published by the Free Software Foundation; either version 2 of
@@ -40,19 +38,13 @@
 #include <string.h>
 #include <errno.h>
 #include <sys/types.h>
-#include <fcntl.h>
 #include <signal.h>
 
 #include "utilvfs.h"
 
-#include "../src/dir.h"
-#include "../src/main.h"
-#ifndef VFS_STANDALONE
-#include "../src/panel.h"
-#include "../src/key.h"		/* Required for the async alarm handler */
-#include "../src/layout.h"	/* For get_panel_widget and get_other_index */
-#include "../src/dialog.h"
-#endif
+#include "../src/panel.h"	/* get_current_panel() */
+#include "../src/layout.h"	/* get_current_type() */
+#include "../src/wtools.h"	/* input_dialog() */
 
 #include "xdirentry.h"
 #include "vfs.h"
@@ -62,12 +54,8 @@
 #   include "tcputil.h"
 #endif
 
-extern int get_other_type (void);
-
 int vfs_timeout = 60; /* VFS timeout in seconds */
-int vfs_flags = 0;    /* Flags */
-
-extern int cd_symlinks; /* Defined in main.c */
+static int vfs_flags = 0;    /* Flags */
 
 /* They keep track of the current directory */
 static vfs *current_vfs = &vfs_local_ops;
@@ -80,9 +68,6 @@ static int current_mday;
 static int current_mon;
 static int current_year;
 
-uid_t vfs_uid = 0;
-gid_t vfs_gid = 0;
-
 /* FIXME: Open files managed by the vfs layer, should be dynamical */
 #define MAX_VFS_FILES 100
 
@@ -109,11 +94,11 @@ get_bucket (void)
 /* vfs_local_ops needs to be the first one */
 static vfs *vfs_list = &vfs_local_ops;
 
-int
+static int
 vfs_register (vfs *vfs)
 {
     if (!vfs)
-	vfs_die("You can not register NULL.");
+	vfs_die("You cannot register NULL.");
     
     if (vfs->init)		/* vfs has own initialization function */
 	if (!(*vfs->init)(vfs))	/* but it failed */
@@ -125,7 +110,7 @@ vfs_register (vfs *vfs)
     return 1;
 }
 
-vfs *
+static vfs *
 vfs_type_from_op (char *path)
 {
     vfs *vfs;
@@ -204,7 +189,7 @@ vfs_split (char *path, char **inpath, char **op)
     vfs *ret;
 
     if (!path)
-	vfs_die("Can not split NULL");
+	vfs_die("Cannot split NULL");
     
     semi = strrchr (path, '#');
     if (!semi || !path_magic(path))
@@ -238,7 +223,7 @@ vfs_split (char *path, char **inpath, char **op)
     return ret;
 }
 
-vfs *
+static vfs *
 vfs_rosplit (char *path)
 {
     char *semi;
@@ -295,17 +280,20 @@ vfs_timeouts ()
     return stamps ? 10 : 0;
 }
 
-void
+static void
 vfs_addstamp (vfs *v, vfsid id, struct vfs_stamping *parent)
 {
     if (v != &vfs_local_ops && id != (vfsid)-1){
-        struct vfs_stamping *stamp, *st1;
+        struct vfs_stamping *stamp;
+	struct vfs_stamping *last_stamp = NULL;
         
-        for (stamp = stamps; stamp != NULL; st1 = stamp, stamp = stamp->next)
+        for (stamp = stamps; stamp != NULL; stamp = stamp->next) {
             if (stamp->v == v && stamp->id == id){
 		gettimeofday(&(stamp->time), NULL);
                 return;
 	    }
+	    last_stamp = stamp;
+	}
         stamp = g_new (struct vfs_stamping, 1);
         stamp->v = v;
         stamp->id = id;
@@ -325,10 +313,13 @@ vfs_addstamp (vfs *v, vfsid id, struct vfs_stamping *parent)
         gettimeofday (&(stamp->time), NULL);
         stamp->next = 0;
 
-	if (stamps)
-	    st1->next = stamp;
-	else
+	if (stamps) {
+	    /* Add to the end */
+	    last_stamp->next = stamp;
+	} else {
+	    /* Add first element */
     	    stamps = stamp;
+	}
     }
 }
 
@@ -445,7 +436,7 @@ mc_open (const char *filename, int flags, ...)
 #define MC_NAMEOP(name, inarg, callarg) \
   MC_OP (name, inarg, callarg, path = vfs_canon (path); vfs = vfs_type (path);, g_free (path); )
 #define MC_HANDLEOP(name, inarg, callarg) \
-  MC_OP (name, inarg, callarg, if (handle == -1) return -1; vfs = vfs_op (handle);, )
+  MC_OP (name, inarg, callarg, if (handle == -1) return -1; vfs = vfs_op (handle);, ;)
 
 MC_HANDLEOP(read, (int handle, char *buffer, int count), (vfs_info (handle), buffer, count) )
 
@@ -521,9 +512,6 @@ mc_opendir (char *dirname)
     return (DIR *) handlep;
 }
 
-/* This should strip the non needed part of a path name */
-#define vfs_name(x) x
-
 void
 mc_seekdir (DIR *dirp, int offset)
 {
@@ -577,37 +565,69 @@ mc_closedir (DIR *dirp)
     return result; 
 }
 
-MC_NAMEOP   (stat, (char *path, struct stat *buf), (vfs, vfs_name (path), buf))
-MC_NAMEOP   (lstat, (char *path, struct stat *buf), (vfs, vfs_name (path), buf))
-MC_HANDLEOP (fstat, (int handle, struct stat *buf), (vfs_info (handle), buf))
+int mc_stat (const char *filename, struct stat *buf) {
+    vfs *vfs;
+    int result;
+    char *path;
+    path = vfs_canon (filename); vfs = vfs_type (path);
+    result = vfs->stat ? (*vfs->stat) (vfs, path, buf) : -1;
+    g_free (path);
+    if (result == -1)
+	errno = vfs->name ? ferrno (vfs) : E_NOTSUPP;
+    return result;
+}
+
+int mc_lstat (const char *filename, struct stat *buf) {
+    vfs *vfs;
+    int result;
+    char *path;
+    path = vfs_canon (filename); vfs = vfs_type (path);
+    result = vfs->lstat ? (*vfs->lstat) (vfs, path, buf) : -1;
+    g_free (path);
+    if (result == -1)
+	errno = vfs->name ? ferrno (vfs) : E_NOTSUPP;
+    return result;
+}
+
+int mc_fstat (int handle, struct stat *buf) {
+    vfs *vfs;
+    int result;
+
+    if (handle == -1)
+	return -1;
+    vfs = vfs_op (handle);
+    result = vfs->fstat ? (*vfs->fstat) (vfs_info (handle), buf) : -1;
+    if (result == -1)
+	errno = vfs->name ? ferrno (vfs) : E_NOTSUPP;
+    return result;
+}
 
 /*
- * You must g_strdup whatever this function returns, static buffers are in use
+ * You must g_strdup whatever this function returns.
  */
 
-char *
+static const char *
 mc_return_cwd (void)
 {
     char *p;
     struct stat my_stat, my_stat2;
 
     if (!vfs_rosplit (current_dir)){
-	static char buffer[MC_MAXPATHLEN];
-
-	p = get_current_wd (buffer, MC_MAXPATHLEN);
+	p = g_get_current_dir ();
 	if (!p)  /* One of the directories in the path is not readable */
 	    return current_dir;
 
 	/* Otherwise check if it is O.K. to use the current_dir */
-	mc_stat (p, &my_stat);
-	mc_stat (current_dir, &my_stat2);
-	if (my_stat.st_ino != my_stat2.st_ino ||
-	    my_stat.st_dev != my_stat2.st_dev ||
-	    !cd_symlinks){
+	if (!cd_symlinks ||
+	    mc_stat (p, &my_stat) || 
+	    mc_stat (current_dir, &my_stat2) ||
+	    my_stat.st_ino != my_stat2.st_ino ||
+	    my_stat.st_dev != my_stat2.st_dev){
 	    g_free (current_dir);
-	    current_dir = g_strdup (p);
+	    current_dir = p;
 	    return p;
 	} /* Otherwise we return current_dir below */
+	g_free (p);
     } 
     return current_dir;
 }
@@ -615,28 +635,29 @@ mc_return_cwd (void)
 char *
 mc_get_current_wd (char *buffer, int size)
 {
-    char *cwd = mc_return_cwd();
+    const char *cwd = mc_return_cwd();
 
-    strncpy (buffer, cwd, size);
+    strncpy (buffer, cwd, size - 1);
+    buffer [size - 1] = 0;
     return buffer;
 }
 
-MC_NAMEOP (chmod, (char *path, int mode), (vfs, vfs_name (path), mode))
-MC_NAMEOP (chown, (char *path, int owner, int group), (vfs, vfs_name (path), owner, group))
-MC_NAMEOP (utime, (char *path, struct utimbuf *times), (vfs, vfs_name (path), times))
-MC_NAMEOP (readlink, (char *path, char *buf, int bufsiz), (vfs, vfs_name (path), buf, bufsiz))
-MC_NAMEOP (unlink, (char *path), (vfs, vfs_name (path)))
-MC_NAMEOP (symlink, (char *name1, char *path), (vfs, vfs_name (name1), vfs_name (path)))
+MC_NAMEOP (chmod, (char *path, int mode), (vfs, path, mode))
+MC_NAMEOP (chown, (char *path, int owner, int group), (vfs, path, owner, group))
+MC_NAMEOP (utime, (char *path, struct utimbuf *times), (vfs, path, times))
+MC_NAMEOP (readlink, (char *path, char *buf, int bufsiz), (vfs, path, buf, bufsiz))
+MC_NAMEOP (unlink, (char *path), (vfs, path))
+MC_NAMEOP (symlink, (char *name1, char *path), (vfs, name1, path))
 
 #define MC_RENAMEOP(name) \
-int mc_##name (char *name1, char *name2) \
+int mc_##name (const char *fname1, const char *fname2) \
 { \
     vfs *vfs; \
     int result; \
 \
-    name1 = vfs_canon (name1); \
+    char *name2, *name1 = vfs_canon (fname1); \
     vfs = vfs_type (name1); \
-    name2 = vfs_canon (name2); \
+    name2 = vfs_canon (fname2); \
     if (vfs != vfs_type (name2)){ \
     	errno = EXDEV; \
     	g_free (name1); \
@@ -644,7 +665,7 @@ int mc_##name (char *name1, char *name2) \
 	return -1; \
     } \
 \
-    result = vfs->name ? (*vfs->name)(vfs, vfs_name (name1), vfs_name (name2)) : -1; \
+    result = vfs->name ? (*vfs->name)(vfs, name1, name2) : -1; \
     g_free (name1); \
     g_free (name2); \
     if (result == -1) \
@@ -682,7 +703,7 @@ char *
 vfs_canon (const char *path)
 {
     if (!path)
-	vfs_die("Can not canonize NULL");
+	vfs_die("Cannot canonicalize NULL");
 
     /* Tilde expansion */
     if (*path == '~'){ 
@@ -713,11 +734,7 @@ vfs_canon (const char *path)
      */
     {
 	char *result = g_strdup (path);
-        
-	mad_check("(pre-canonicalize)", 0);
 	canonicalize_pathname (result);
-	mad_check("(post-canonicalize)", 0);
-
         return result;
     }
 }
@@ -750,7 +767,6 @@ is_parent (vfs * nvfs, vfsid nvfsid, struct vfs_stamping *parent)
 void
 vfs_add_noncurrent_stamps (vfs * oldvfs, vfsid oldvfsid, struct vfs_stamping *parent)
 {
-#ifndef VFS_STANDALONE
     vfs *nvfs, *n2vfs, *n3vfs;
     vfsid nvfsid, n2vfsid, n3vfsid;
     struct vfs_stamping *par, *stamp;
@@ -826,9 +842,6 @@ vfs_add_noncurrent_stamps (vfs * oldvfs, vfsid oldvfsid, struct vfs_stamping *pa
 		vfs_addstamp (stamp->v, stamp->id, stamp->parent);
 	}
     }
-#else
-    vfs_addstamp (oldvfs, oldvfsid, parent);
-#endif
 }
 
 static void
@@ -847,7 +860,6 @@ vfs_stamp_path (char *path)
     vfs_rm_parents (par);
 }
 
-#ifndef VFS_STANDALONE
 void
 vfs_add_current_stamps (void)
 {
@@ -863,44 +875,57 @@ vfs_add_current_stamps (void)
 	    vfs_stamp_path (opanel->cwd);
     }
 }
-#endif
 
-/* This function is really broken */
+/*
+ * VFS chdir.
+ * Return 0 on success, -1 on failure.
+ */
 int
 mc_chdir (char *path)
 {
-    char *a, *b;
-    int result;
-    char *p = NULL;
-    vfs *oldvfs;
-    vfsid oldvfsid;
+    char *new_dir, *new_dir_copy;
+    vfs *old_vfs, *new_vfs;
+    vfsid old_vfsid;
     struct vfs_stamping *parent;
+    int result;
 
-    a = current_dir; /* Save a copy for case of failure */
-    current_dir = vfs_canon (path);
-    current_vfs = vfs_type (current_dir);
-    b = g_strdup (current_dir);
-    result = (*current_vfs->chdir) ?  (*current_vfs->chdir)(current_vfs, vfs_name (b)) : -1;
-    g_free (b);
-    if (result == -1){
-	errno = ferrno (current_vfs);
-    	g_free (current_dir);
-    	current_vfs = vfs_type (a);
-    	current_dir = a;
-    } else {
-        oldvfs = vfs_type (a);
-	oldvfsid = vfs_ncs_getid (oldvfs, a, &parent);
-	g_free (a);
-        vfs_add_noncurrent_stamps (oldvfs, oldvfsid, parent);
-	vfs_rm_parents (parent);
+    new_dir = vfs_canon (path);
+    new_vfs = vfs_type (new_dir);
+    if (!new_vfs->chdir)
+	vfs_die ("No chdir function defined");
+
+    /* new_vfs->chdir can write to the second argument, use a copy */
+    new_dir_copy = g_strdup (new_dir);
+    result = (*new_vfs->chdir) (new_vfs, new_dir_copy);
+    g_free (new_dir_copy);
+
+    if (result == -1) {
+	errno = ferrno (new_vfs);
+	g_free (new_dir);
+	return -1;
     }
 
-    if (*current_dir){
+    old_vfsid = vfs_ncs_getid (current_vfs, current_dir, &parent);
+    old_vfs = current_vfs;
+
+    /* Actually change directory */
+    g_free (current_dir);
+    current_dir = new_dir;
+    current_vfs = new_vfs;
+
+    /* This function uses the new current_dir implicitly */
+    vfs_add_noncurrent_stamps (old_vfs, old_vfsid, parent);
+    vfs_rm_parents (parent);
+
+    /* Sometimes we assume no trailing slash on cwd */
+    if (*current_dir) {
+	char *p;
 	p = strchr (current_dir, 0) - 1;
 	if (*p == PATH_SEP && p > current_dir)
-	    *p = 0; /* Sometimes we assume no trailing slash on cwd */
+	    *p = 0;
     }
-    return result;
+
+    return 0;
 }
 
 int
@@ -909,27 +934,6 @@ vfs_current_is_local (void)
     return current_vfs == &vfs_local_ops;
 }
 
-#if 0
-/* External world should not do differences between VFS-s */
-int
-vfs_current_is_extfs (void)
-{
-    return current_vfs == &vfs_extfs_ops;
-}
-
-int
-vfs_current_is_tarfs (void)
-{
-    return current_vfs == &vfs_tarfs_ops;
-}
-
-int
-vfs_current_is_cpiofs (void)
-{
-    return current_vfs == &vfs_cpiofs_ops;
-}
-#endif
-
 int
 vfs_file_is_local (const char *file)
 {
@@ -986,9 +990,9 @@ static void vfs_setup_wd (void)
         vfs_die ("Current dir too long.\n");
 }
 
-MC_NAMEOP (mkdir, (char *path, mode_t mode), (vfs, vfs_name (path), mode))
-MC_NAMEOP (rmdir, (char *path), (vfs, vfs_name (path)))
-MC_NAMEOP (mknod, (char *path, int mode, int dev), (vfs, vfs_name (path), mode, dev))
+MC_NAMEOP (mkdir, (char *path, mode_t mode), (vfs, path, mode))
+MC_NAMEOP (rmdir, (char *path), (vfs, path))
+MC_NAMEOP (mknod, (char *path, int mode, int dev), (vfs, path, mode, dev))
 
 #ifdef HAVE_MMAP
 static struct mc_mmapping {
@@ -1052,12 +1056,25 @@ mc_def_getlocalcopy (vfs *vfs, char *filename)
     int fdin, fdout, i;
     char buffer[8192];
     struct stat mystat;
+    char *ext = NULL;
+    char *ptr;
 
     fdin = mc_open (filename, O_RDONLY);
     if (fdin == -1)
         return NULL;
-    tmp = g_tempnam (NULL, "mclocalcopy");
-    fdout = open (tmp, O_CREAT|O_WRONLY|O_TRUNC|O_EXCL, 0600);
+
+    /* Try to preserve existing extension */
+    for (ptr = filename + strlen(filename) - 1; ptr >= filename; ptr--) {
+	if (*ptr == '.') {
+	    ext = ptr;
+	    break;
+	}
+
+	if (!isalnum((unsigned char) *ptr))
+	    break;
+    }
+
+    fdout = mc_mkstemps (&tmp, "mclocalcopy", ext);
     if (fdout == -1)
 	goto fail;
     while ((i = mc_read (fdin, buffer, sizeof (buffer))) > 0){
@@ -1092,8 +1109,8 @@ mc_getlocalcopy (const char *pathname)
     char *path = vfs_canon (pathname);
     vfs  *vfs  = vfs_type (path);    
 
-    result = vfs->getlocalcopy ? (*vfs->getlocalcopy)(vfs, vfs_name (path)) :
-                                 mc_def_getlocalcopy (vfs, vfs_name (path));
+    result = vfs->getlocalcopy ? (*vfs->getlocalcopy)(vfs, path) :
+                                 mc_def_getlocalcopy (vfs, path);
     g_free (path);
     if (!result)
 	errno = ferrno (vfs);
@@ -1151,8 +1168,8 @@ mc_ungetlocalcopy (const char *pathname, char *local, int has_changed)
     vfs *vfs = vfs_type (path);
 
     return_value = vfs->ungetlocalcopy ? 
-            (*vfs->ungetlocalcopy)(vfs, vfs_name (path), local, has_changed) :
-            mc_def_ungetlocalcopy (vfs, vfs_name (path), local, has_changed);
+            (*vfs->ungetlocalcopy)(vfs, path, local, has_changed) :
+            mc_def_ungetlocalcopy (vfs, path, local, has_changed);
     g_free (path);
     return return_value;
 }
@@ -1160,7 +1177,7 @@ mc_ungetlocalcopy (const char *pathname, char *local, int has_changed)
 /*
  * Hmm, as timeout is minute or so, do we need to care about usecs?
  */
-inline int
+static inline int
 timeoutcmp (struct timeval *t1, struct timeval *t2)
 {
     return ((t1->tv_sec < t2->tv_sec)
@@ -1222,13 +1239,15 @@ vfs_init (void)
 #ifdef USE_NETCODE
     tcp_init();
     vfs_register (&vfs_ftpfs_ops);
+    vfs_register (&vfs_fish_ops);
 #ifdef WITH_SMBFS
     vfs_register (&vfs_smbfs_ops);
-#endif
+#endif /* WITH_SMBFS */
+#ifdef WITH_MCFS
     vfs_register (&vfs_mcfs_ops);
-#endif
+#endif /* WITH_SMBFS */
+#endif /* USE_NETCODE */
 
-    vfs_register (&vfs_fish_ops);
     vfs_register (&vfs_extfs_ops);
     vfs_register (&vfs_sfs_ops);
     vfs_register (&vfs_tarfs_ops);
@@ -1236,40 +1255,11 @@ vfs_init (void)
 
 #ifdef USE_EXT2FSLIB
     vfs_register (&vfs_undelfs_ops);
-#endif
+#endif /* USE_EXT2FSLIB */
 
     vfs_setup_wd ();
 }
 
-void
-vfs_free_resources (char *path)
-{
-    vfs *vfs;
-    vfsid vid;
-    struct vfs_stamping *parent;
-    
-    vfs = vfs_type (path);
-    vid = vfs_ncs_getid (vfs, path, &parent);
-    if (vid != (vfsid) -1)
-	(*vfs->free)(vid);
-    vfs_rm_parents (parent);
-}
-
-#if 0
-/* Shutdown a vfs given a path name */
-void
-vfs_shut_path (char *p)
-{
-    vfs *the_vfs;
-    struct vfs_stamping *par;
-
-    the_vfs = vfs_type (p);
-    vfs_ncs_getid (the_vfs, p, &par);
-    (*par->v->free)(par->id);
-    vfs_rm_parents (par);
-}
-#endif
-
 void
 vfs_shut (void)
 {
@@ -1320,6 +1310,8 @@ vfs_split_text (char *p)
     char *original = p;
     int  numcols;
 
+    memset (columns, 0, sizeof (columns));
+
     for (numcols = 0; *p && numcols < MAXCOLS; numcols++){
 	while (*p == ' ' || *p == '\r' || *p == '\n'){
 	    *p = 0;
@@ -1336,46 +1328,68 @@ vfs_split_text (char *p)
 static int
 is_num (int idx)
 {
-    if (!columns [idx] || columns [idx][0] < '0' || columns [idx][0] > '9')
+    char *column = columns[idx];
+
+    if (!column || column[0] < '0' || column[0] > '9')
 	return 0;
+
     return 1;
 }
 
+/* Return 1 for MM-DD-YY and MM-DD-YYYY */
 static int
-is_dos_date(char *str)
+is_dos_date (const char *str)
 {
-    if (strlen(str) == 8 && str[2] == str[5] && strchr("\\-/", (int)str[2]) != NULL)
-	return (1);
+    int len;
 
-    return (0);
+    if (!str)
+	return 0;
+
+    len = strlen (str);
+    if (len != 8 && len != 10)
+	return 0;
+
+    if (str[2] != str[5])
+	return 0;
+
+    if (!strchr ("\\-/", (int) str[2]))
+	return 0;
+
+    return 1;
 }
 
 static int
 is_week (char *str, struct tm *tim)
 {
-    static char *week = "SunMonTueWedThuFriSat";
+    static const char *week = "SunMonTueWedThuFriSat";
     char *pos;
 
-    if((pos=strstr(week, str)) != NULL){
-        if(tim != NULL)
-	    tim->tm_wday = (pos - week)/3;
-	return (1);
+    if (!str)
+	return 0;
+
+    if ((pos = strstr (week, str)) != NULL) {
+	if (tim != NULL)
+	    tim->tm_wday = (pos - week) / 3;
+	return 1;
     }
-    return (0);    
+    return 0;
 }
 
 static int
 is_month (char *str, struct tm *tim)
 {
-    static char *month = "JanFebMarAprMayJunJulAugSepOctNovDec";
+    static const char *month = "JanFebMarAprMayJunJulAugSepOctNovDec";
     char *pos;
-    
-    if((pos=strstr(month, str)) != NULL){
-        if(tim != NULL)
-	    tim->tm_mon = (pos - month)/3;
-	return (1);
+
+    if (!str)
+	return 0;
+
+    if ((pos = strstr (month, str)) != NULL) {
+	if (tim != NULL)
+	    tim->tm_mon = (pos - month) / 3;
+	return 1;
     }
-    return (0);
+    return 0;
 }
 
 static int
@@ -1383,41 +1397,47 @@ is_time (char *str, struct tm *tim)
 {
     char *p, *p2;
 
-    if ((p=strchr(str, ':')) && (p2=strrchr(str, ':'))) {
+    if (!str)
+	return 0;
+
+    if ((p = strchr (str, ':')) && (p2 = strrchr (str, ':'))) {
 	if (p != p2) {
-    	    if (sscanf (str, "%2d:%2d:%2d", &tim->tm_hour, &tim->tm_min, &tim->tm_sec) != 3)
-		return (0);
-	}
-	else {
+	    if (sscanf
+		(str, "%2d:%2d:%2d", &tim->tm_hour, &tim->tm_min,
+		 &tim->tm_sec) != 3)
+		return 0;
+	} else {
 	    if (sscanf (str, "%2d:%2d", &tim->tm_hour, &tim->tm_min) != 2)
-		return (0);
+		return 0;
 	}
-    }
-    else 
-        return (0);
-    
-    return (1);
+    } else
+	return 0;
+
+    return 1;
 }
 
-static int is_year(char *str, struct tm *tim)
+static int is_year (char *str, struct tm *tim)
 {
     long year;
 
-    if (strchr(str,':'))
-        return (0);
+    if (!str)
+	return 0;
 
-    if (strlen(str)!=4)
-        return (0);
+    if (strchr (str, ':'))
+	return 0;
 
-    if (sscanf(str, "%ld", &year) != 1)
-        return (0);
+    if (strlen (str) != 4)
+	return 0;
+
+    if (sscanf (str, "%ld", &year) != 1)
+	return 0;
 
     if (year < 1900 || year > 3000)
-        return (0);
+	return 0;
 
     tim->tm_year = (int) (year - 1900);
 
-    return (1);
+    return 1;
 }
 
 /*
@@ -1429,19 +1449,28 @@ static int is_year(char *str, struct tm *tim)
 int
 vfs_parse_filetype (char c)
 {
-    switch (c){
-        case 'd': return S_IFDIR; 
-        case 'b': return S_IFBLK;
-        case 'c': return S_IFCHR;
-        case 'l': return S_IFLNK;
-        case 's':
-#ifdef IS_IFSOCK /* And if not, we fall through to IFIFO, which is pretty close */
-	          return S_IFSOCK;
+    switch (c) {
+	case 'd': return S_IFDIR; 
+	case 'b': return S_IFBLK;
+	case 'c': return S_IFCHR;
+	case 'l': return S_IFLNK;
+	case 's': /* Socket */
+#ifdef S_IFSOCK
+		return S_IFSOCK;
+#else
+		/* If not supported, we fall through to IFIFO */
+		return S_IFIFO;
 #endif
-        case 'p': return S_IFIFO;
-        case 'm': case 'n':		/* Don't know what these are :-) */
-        case '-': case '?': return S_IFREG;
-        default: return -1;
+	case 'D': /* Solaris door */
+#ifdef S_IFDOOR
+		return S_IFDOOR;
+#else
+		return S_IFIFO;
+#endif
+	case 'p': return S_IFIFO;
+	case 'm': case 'n':		/* Don't know what these are :-) */
+	case '-': case '?': return S_IFREG;
+	default: return -1;
     }
 }
 
@@ -1503,92 +1532,97 @@ int vfs_parse_filemode (const char *p)
     return res;
 }
 
-int vfs_parse_filedate(int idx, time_t *t)
-{	/* This thing parses from idx in columns[] array */
-
+/* This function parses from idx in the columns[] array */
+int
+vfs_parse_filedate (int idx, time_t *t)
+{
     char *p;
     struct tm tim;
     int d[3];
-    int	got_year = 0;
+    int got_year = 0;
 
     /* Let's setup default time values */
     tim.tm_year = current_year;
-    tim.tm_mon  = current_mon;
+    tim.tm_mon = current_mon;
     tim.tm_mday = current_mday;
     tim.tm_hour = 0;
-    tim.tm_min  = 0;
-    tim.tm_sec  = 0;
-    tim.tm_isdst = -1; /* Let mktime() try to guess correct dst offset */
-    
-    p = columns [idx++];
-    
+    tim.tm_min = 0;
+    tim.tm_sec = 0;
+    tim.tm_isdst = -1;		/* Let mktime() try to guess correct dst offset */
+
+    p = columns[idx++];
+
     /* We eat weekday name in case of extfs */
-    if(is_week(p, &tim))
-	p = columns [idx++];
+    if (is_week (p, &tim))
+	p = columns[idx++];
 
     /* Month name */
-    if(is_month(p, &tim)){
+    if (is_month (p, &tim)) {
 	/* And we expect, it followed by day number */
 	if (is_num (idx))
-    	    tim.tm_mday = (int)atol (columns [idx++]);
+	    tim.tm_mday = (int) atol (columns[idx++]);
 	else
-	    return 0; /* No day */
+	    return 0;		/* No day */
 
     } else {
-        /* We usually expect:
-           Mon DD hh:mm
-           Mon DD  YYYY
+	/* We usually expect:
+	   Mon DD hh:mm
+	   Mon DD  YYYY
 	   But in case of extfs we allow these date formats:
-           Mon DD YYYY hh:mm
+	   Mon DD YYYY hh:mm
 	   Mon DD hh:mm YYYY
 	   Wek Mon DD hh:mm:ss YYYY
-           MM-DD-YY hh:mm
-           where Mon is Jan-Dec, DD, MM, YY two digit day, month, year,
-           YYYY four digit year, hh, mm, ss two digit hour, minute or second. */
+	   MM-DD-YY hh:mm
+	   where Mon is Jan-Dec, DD, MM, YY two digit day, month, year,
+	   YYYY four digit year, hh, mm, ss two digit hour, minute or second. */
 
-	/* Here just this special case with MM-DD-YY */
-        if (is_dos_date(p)){
-            p[2] = p[5] = '-';
-	    
-	    if(sscanf(p, "%2d-%2d-%2d", &d[0], &d[1], &d[2]) == 3){
-	    /*  We expect to get:
-		1. MM-DD-YY
-		2. DD-MM-YY
-		3. YY-MM-DD
-		4. YY-DD-MM  */
-		
-		/* Hmm... maybe, next time :)*/
-		
-		/* At last, MM-DD-YY */
-		d[0]--; /* Months are zerobased */
-		/* Y2K madness */
-		if(d[2] < 70)
-		    d[2] += 100;
+	/* Special case with MM-DD-YY or MM-DD-YYYY */
+	if (is_dos_date (p)) {
+	    p[2] = p[5] = '-';
 
-        	tim.tm_mon  = d[0];
-        	tim.tm_mday = d[1];
-        	tim.tm_year = d[2];
+	    if (sscanf (p, "%2d-%2d-%d", &d[0], &d[1], &d[2]) == 3) {
+		/* Months are zero based */
+		if (d[0] > 0)
+		    d[0]--;
+
+		if (d[2] > 1900) {
+		    d[2] -= 1900;
+		} else {
+		    /* Y2K madness */
+		    if (d[2] < 70)
+			d[2] += 100;
+		}
+
+		tim.tm_mon = d[0];
+		tim.tm_mday = d[1];
+		tim.tm_year = d[2];
 		got_year = 1;
 	    } else
-		return 0; /* sscanf failed */
-        } else
-            return 0; /* unsupported format */
+		return 0;	/* sscanf failed */
+	} else
+	    return 0;		/* unsupported format */
     }
 
     /* Here we expect to find time and/or year */
-    
+
     if (is_num (idx)) {
-        if(is_time(columns[idx], &tim) || (got_year = is_year(columns[idx], &tim))) {
+	if (is_time (columns[idx], &tim)
+	    || (got_year = is_year (columns[idx], &tim))) {
 	    idx++;
 
-	/* This is a special case for ctime() or Mon DD YYYY hh:mm */
-	if(is_num (idx) && (columns[idx+1][0]) &&
-	    ((got_year |= is_year(columns[idx], &tim)) || is_time(columns[idx], &tim)))
-		idx++; /* time & year or reverse */
-	} /* only time or date */
-    }
-    else 
-        return 0; /* Nor time or date */
+	    /* This is a special case for ctime() or Mon DD YYYY hh:mm */
+	    if (is_num (idx) && (columns[idx + 1][0])) {
+		if (got_year) {
+		    if (is_time (columns[idx], &tim))
+			idx++;	/* time also */
+		} else {
+		    if ((got_year = is_year (columns[idx], &tim)))
+			idx++;	/* year also */
+		}
+	    }
+	}			/* only time or date */
+    } else
+	return 0;		/* Nor time or date */
 
     /*
      * If the date is less than 6 months in the past, it is shown without year
@@ -1596,14 +1630,13 @@ int vfs_parse_filedate(int idx, time_t *t)
      * This does not check for years before 1900 ... I don't know, how
      * to represent them at all
      */
-    if (!got_year &&
-	current_mon < 6 && current_mon < tim.tm_mon && 
-	tim.tm_mon - current_mon >= 6)
+    if (!got_year && current_mon < 6 && current_mon < tim.tm_mon
+	&& tim.tm_mon - current_mon >= 6)
 
 	tim.tm_year--;
 
-    if ((*t = mktime(&tim)) < 0)
-        *t = 0;
+    if ((*t = mktime (&tim)) < 0)
+	*t = 0;
     return idx;
 }
 
@@ -1612,13 +1645,13 @@ vfs_parse_ls_lga (const char *p, struct stat *s, char **filename, char **linknam
 {
     int idx, idx2, num_cols;
     int i;
-    char *p_copy;
+    char *p_copy = NULL;
     char *t = NULL;
+    const char *line = p;
 
     if (strncmp (p, "total", 5) == 0)
         return 0;
 
-    p_copy = g_strdup(p);
     if ((i = vfs_parse_filetype(*(p++))) == -1)
         goto error;
 
@@ -1645,12 +1678,11 @@ vfs_parse_ls_lga (const char *p, struct stat *s, char **filename, char **linknam
             p++;
     }
 
-    g_free(p_copy);
     p_copy = g_strdup(p);
     num_cols = vfs_split_text (p_copy);
 
     s->st_nlink = atol (columns [0]);
-    if (s->st_nlink < 0)
+    if (s->st_nlink <= 0)
         goto error;
 
     if (!is_num (1))
@@ -1709,11 +1741,7 @@ vfs_parse_ls_lga (const char *p, struct stat *s, char **filename, char **linknam
         goto error;
     /* Use resulting time value */
     s->st_atime = s->st_ctime = s->st_mtime;
-#if 0
-    /* These variables must be initialized by vfs_s_new_inode () */
-    s->st_dev = 0;
-    s->st_ino = 0;
-#endif
+    /* s->st_dev and s->st_ino must be initialized by vfs_s_new_inode () */
 #ifdef HAVE_ST_BLKSIZE
     s->st_blksize = 512;
 #endif
@@ -1770,9 +1798,9 @@ error:
       static int errorcount = 0;
 
       if (++errorcount < 5) {
-	message_1s (1, _("Could not parse:"), p_copy);
+	message_1s (1, _("Cannot parse:"), (p_copy && *p_copy) ? p_copy : line);
       } else if (errorcount == 5)
-	message_1s (1, _("More parsing errors will be ignored."), _("(sorry)"));
+	message_1s (1, _("Error"), _("More parsing errors will be ignored."));
     }
 
     g_free (p_copy);
@@ -1780,37 +1808,35 @@ error:
 }
 
 void
-vfs_die (char *m)
+vfs_die (const char *m)
 {
     message_1s (1, _("Internal error:"), m);
     exit (1);
 }
 
 void
-vfs_print_stats (char *fs_name, char *action, char *file_name, int have, int need)
+vfs_print_stats (const char *fs_name, const char *action, const char *file_name, off_t have, off_t need)
 {
     static char *i18n_percent_transf_format = NULL, *i18n_transf_format = NULL;
-    
+
     if (i18n_percent_transf_format == NULL) {
-        i18n_percent_transf_format = _("%s: %s: %s %3d%% (%ld bytes transfered)");
-	i18n_transf_format = _("%s: %s: %s %ld bytes transfered");
-	}
+	i18n_percent_transf_format = _("%s: %s: %s %3d%% (%lu bytes transferred)");
+	i18n_transf_format = _("%s: %s: %s %lu bytes transferred");
+    }
 
     if (need) 
-        print_vfs_message (i18n_percent_transf_format, 
-			   fs_name, action, file_name, have*100/need, have);
+	print_vfs_message (i18n_percent_transf_format, fs_name, action,
+			   file_name, (int)((double)have*100/need), (unsigned long) have);
     else
-        print_vfs_message (i18n_transf_format,
-			   fs_name, action, file_name, have);
+	print_vfs_message (i18n_transf_format,
+			   fs_name, action, file_name, (unsigned long) have);
 }
 
-#ifndef VFS_STANDALONE
 char *
 vfs_get_password (char *msg)
 {
-    return (char *) input_dialog (msg, _("Password:"), "");
+    return (char *) input_dialog (msg, _("Password:"), INPUT_PASSWORD);
 }
-#endif
 
 /*
  * Returns vfs path corresponding to given url. If passed string is
